
# Hacky quick-fix allowing importing of cadence module. 
# Should be removed when this is expanded into a "real" testing suite
import sys
sys.path.append("..")

# Sample Test passing with nose and pytest

def test_pass():
    assert True, "dummy sample test"

def test_import():
    import cadence as cd
    assert True, "Ensure that Cadence can be imported"

def test_scan():
    import cadence as cd

    milton = "OF Mans First Disobedience, and the Fruit"\
    "Of that Forbidden Tree, whose mortal tast"\
    "Brought Death into the World, and all our woe"

    txtdf = cd.scan(milton)
    assert txtdf is not None, "Ensure that Cadence can scan without Exceptions"



from ...imports import *
from .deptree import *
from .metricaltree import *
STRESS_STRS=set("`'")
ALLOW_AMBIG_MTREE=False


def recurse_tree(tree,node_i=0,path=[]):
    o=[]
    if not hasattr(tree,'node_i'):
        tree.node_i=node_i
        node_i+=1

    # path+=[f'{tree.node_i}-{tree.label}']
    if not tree.is_leaf():
        path+=[f'{tree.node_i}-{tree.label}']
        for x in tree.children:
            o+=recurse_tree(x,node_i=node_i,path=list(path))
    else:
        o+=[tuple(path)]
        path=[]
    return o


def get_dtree_str(sent):
    o=[]
    for dep in sent.dependencies:
        w1,rel,w2=dep
        ostr=f'{rel}({w1.text}-{w1.id}, {w2.text}-{w2.id})'
        o+=[ostr]
    return '\n'.join(o)


def get_ctree_str(sent):
    return str(sent.constituency)

def get_treeparse_str(sent):
    ctree=get_ctree_str(sent)
    dtree=get_dtree_str(sent)
    return f'{ctree}\n\n{dtree}'

def get_mtree(sent,**kwargs):
    try:
        tree_str=get_treeparse_str(sent)
        depobj=DependencyTree.fromstring(tree_str)
        metrobj=CadenceMetricalTree.convert(depobj)
        return metrobj
    #except (IndexError,KeyError,AttributeError,ValueError) as e:
    except AssertionError as e:
        print('!!',e)
        pass


def find_phrasal_heads(self):
    for subtree in self:
        if type(subtree)!=str:
            numsubs=len(subtree)
            firstsub,lastsub=subtree[0],subtree[-1]
            if numsubs>1 and not subtree._preterm and lastsub._preterm:
                if lastsub._word_tok: lastsub._phrasal_head=1
                # print('phrasal head', str(subtree))
                for sub in subtree[:-1]:
                    if sub._word_tok: sub._phrasal_head=0
                    # print('not phrasal head', str(sub))
            find_phrasal_heads(subtree)



class CadenceMetricalTree(MetricalTree):
    def __init__(self,
            node,
            children,
            dep=None,
            lstress=0,
            pstress=np.nan,
            stress=np.nan,
            allow_ambig=ALLOW_AMBIG_MTREE,
            **kwargs):

        self._lstress = lstress
        self._pstress = pstress
        self._stress = stress
        super(MetricalTree, self).__init__(node, children, dep)
        self.set_label()
        self._phrasal_head=np.nan
        self._phrasal_stress_peak=np.nan
        self._phrasal_stress_valley=np.nan
        self._pstrength=np.nan
        self._stress_num_binary=1.0
        self._stress_num=1.0
        self._seg = None
        self._nsyll = np.nan
        self._nstress = np.nan
        self._word_tok=''
        

        if self._preterm:
            word_str=str(self[0])
            word_tok=to_token(word_str)
            if word_tok:
                sylls_df=get_syllable_df(word_str,**kwargs)
                if len(sylls_df):
                    self._num_variants=sylls_df.word_ipa_i.max()
                    self._is_ambig=self._num_variants>1
                    self._word_tok=word_tok
                    
                    # go for least stressed possible
                    sylgrps = [g for i,g in sylls_df.groupby('word_ipa_i')]
                    sylgrps.sort(key=lambda g: g.prom_stress.sum())
                    sylls_df1=sylgrps[0]
                    
                    self._nsyll=sylls_df1.syll_i.max()
                    self._seg='.'.join(sylls_df1.syll_ipa)
                    self._nstress=len([x for x in sylls_df1.prom_stress if x>0])
                    
                    # ambig?
                    self._stress_num=sylls_df1.prom_stress.max()
                    if allow_ambig and self._nsyll==1 and self._is_ambig:
                        self._stress_num=.5
                        self._lstress=self._stress_num - 1.0
                        self._stress_num_binary = 1.0
                    else:
                        self._stress_num_binary = 1.0 if self._stress_num>0 else 0.0
                        self._lstress=self._stress_num_binary - 1.0
            else:
                self._stress_num_binary=np.nan
                self._stress_num=np.nan
                self._word_tok=''


    def lstress(self): return self._lstress
    def pstress(self): return self._pstress
    def pstress_pos(self): return -(self._pstress-1)
    def stress(self): return self._stress
    def stress_pos(self): return -(self._stress-1)
    def seg(self): return self._seg if self._seg is not None else []
    def nseg(self): return len(self._seg) if self._seg is not None else np.nan
    def nsyll(self): return self._nsyll
    def nstress(self): return self._nstress


    # =====================================================================
    # Set the lexical stress of the node
    def set_lstress(self,allow_ambig=ALLOW_AMBIG_MTREE):
        # self.gen_word_info()

        if self._preterm:
            if self[0].lower() in super(MetricalTree, self)._contractables:
                self._lstress = np.nan
            elif self._cat in super(MetricalTree, self)._punctTags:
                self._lstress = np.nan
            elif self._cat in MetricalTree._unstressedTags:
                self._lstress = -1
            elif allow_ambig and self._cat in MetricalTree._ambiguousTags:
                self._lstress = -.5
            elif self._dep in MetricalTree._unstressedDeps:
                self._lstress = -1
            elif allow_ambig and self._dep in MetricalTree._ambiguousDeps:
                self._lstress = -.5
            elif allow_ambig and self._nsyll==1 and self._is_ambig:
                self._lstress = -.5
            elif allow_ambig:
                self._lstress = self._stress_num - 1.0
            else:
                self._lstress = self._stress_num_binary - 1.0

            # print(self,self._lstress)
        else:
            for child in self:
                child.set_lstress()
        self.set_label()

    def set_phrasal_peaks(self):
        preterms=list(self.preterminals())
        for i2,w2 in enumerate(preterms):
            w1=preterms[i2-1] if i2 else None
            w3=preterms[i2+1] if (i2+1)<len(preterms) else None
            if np.isnan(w2._pstress): continue
            if w1 and not np.isnan(w1._pstress) and w1._pstress < w2._pstress:
                w1._phrasal_stress_valley=1
                w1._phrasal_stress_peak=0
                w1._pstrength=0.0

                w2._phrasal_stress_peak=1
                w2._phrasal_stress_valley=0
                w2._pstrength=1.0
            if w3 and not np.isnan(w3._pstress) and w3._pstress < w2._pstress:
                w3._phrasal_stress_valley=1
                w3._phrasal_stress_peak=0
                w3._pstrength=0.0

                w2._phrasal_stress_peak=1
                w2._phrasal_stress_valley=0
                w2._pstrength=1.0

    def get_stats(self, arto=False, **kwargs):
        #self.gen_word_info()
        try:
            data = defaultdict(list)
            self.set_lstress()
            self.set_pstress()
            self.set_stress()
            find_phrasal_heads(self)
            self.set_phrasal_peaks()
            
            j = 0
            preterms = list(self.preterminals())
            min1p = float(min([preterm.pstress() for preterm in preterms if not np.isnan(preterm.pstress())]))
            max1p = max([preterm.pstress() for preterm in preterms if not np.isnan(preterm.pstress())]) - min1p
            min1s = float(min([preterm.stress() for preterm in preterms if not np.isnan(preterm.stress())]))
            max1s = max([preterm.stress() for preterm in preterms if not np.isnan(preterm.stress())]) - min1s
            # sent1 = ' '.join([preterm[0] for preterm in preterms])
            # sentlen1 = len(preterms)

            for j,preterm in enumerate(preterms):
                data['word_i'].append(j+1)
                data['word_str'].append(preterm[0])
                # data['mtree_pstress'].append(preterm.pstress())
                data['prom_lstress'].append(preterm.lstress()+1)
                data['prom_pstress'].append((preterm.pstress()-min1p)/max1p if max1p else np.nan)
                data['prom_tstress'].append((preterm.stress()-min1s)/max1s if max1s else np.nan)
                data['prom_pstrength'].append(preterm._pstrength)
                data['mtree_ishead'].append(preterm._phrasal_head)
            return pd.DataFrame(data)
        except ValueError as e:
            # eprint('!!',e,'!!')
            return pd.DataFrame()


from ..imports import *



def iter_combos(txtdf,num_proc=1,**kwargs):
    try:
        txtdf=resetindex(txtdf)
        # txtdf=txtdf[txtdf.word_tok!=""]
        for dfi,dfcombo in enumerate(apply_combos(txtdf, 'word_i', 'word_ipa_i',**kwargs)):
            yield dfcombo.assign(combo_i=dfi+1)
    except KeyError:
        yield pd.DataFrame()

def apply_combos(df,group1,group2,**kwargs):
    # combo of indices?
    combo_opts = [
        [x for ii,x in grp.groupby(group2)]
        for i,grp in df.groupby(group1)
    ]

    # poss
    for combo in product(*combo_opts):
        if not len(combo): continue
        odf=pd.concat(combo)
        odf['slot_i']=[i+1 for i in range(len(odf))]
        yield odf.set_index('slot_i')



def tokenize_sentwords_ll(words_ld,**kwargs):
    if not len(words_ld): return []
    listd=defaultdict(list)
    for word_d in words_ld:
        listd[word_d['sent_i']].append(word_d['word_str'].strip())
    olist=[sentl for sent_i,sentl in sorted(listd.items())]
    return olist


def tokenize_sentwords_ll_from_df(words_df,**kwargs):
    if not len(words_df): return []
    return [
        [w.strip() for w in sdf.sort_values('word_i').word_str]
        for si,sdf in sorted(words_df.groupby('sent_i'))
    ]




def tokenize_sents_txt(txt,**y):
    sents=nltk.sent_tokenize(txt)
    lastoffset=0
    osents=[]
    
    for sent in sents:
        offset = txt.find(sent,lastoffset)
        newpref=txt[lastoffset:offset]
        #print([newpref,offset,lastoffset,sent])
        lastoffset=offset + len(sent)
        newsent=newpref + sent
        osents.append(newsent)
    return osents

# def tokenize_words_txt(txt):
#     l=tokenize_nice2(txt)
#     o=[]
#     o0=''
#     for x in l:
#         if zero_punc(x):
#             o+=[o0 + x]
#             o0=''
#         elif o:
#             o[-1]+=x
#         else:
#             o0+=x
#     return o

def tokenize_words_txt(txt):
    l=tokenize_agnostic(txt)
    o=[]
    x0=''
    for x in l:
        if not x.strip():
            x0+=x
        else:
            o+=[x0 + x]
            x0=''
        # if o and not x.strip():# and not o[-1].strip():
        #     o[-1]+=x
        # else:
        #     o+=[x]
    return o

def tokenize_sentwords_iter(
        txt,
        sents=None,
        lang=DEFAULT_LANG,
        sep_line=SEP_LINE,
        sep_para=SEP_STANZA,
        seps_phrase=SEPS_PHRASE,
        progress=True,
        para_i=None,
        **kwargs):
    char_i=0
    line_i=1
    linepart_i=1
    linepart_ii=0
    start_offset=0
    if sents is None: sents=tokenize_sents_txt(txt)
    for sent_i, sent in enumerate(sents):
        tokens=tokenize_words_txt(sent)
        for tok_i,word_str in enumerate(tokens):
            # word_tok=to_token(word_str)
            if sep_line in word_str: line_i+=1
            is_punc=int(not any(x.isalpha() for x in word_str))
            odx_word=dict(
                # para_i=para_i,
                **(dict(para_i=para_i) if para_i is not None else {}),
                sent_i=sent_i+1,
                sentpart_i=linepart_i,
                line_i=line_i,
                word_i=tok_i+1,
                word_str=word_str,
                # word_tok=word_tok,
                word_ispunc=is_punc
            )
            yield odx_word
            if set(word_str)&set(seps_phrase): linepart_i+=1


def tokenize_paras_ld(txt,sep_para=SEP_PARA, **kwargs):
    txt=clean_text(txt)
    paras_txt=txt.split(sep_para)
    num_paras=len(paras_txt)
    ld=[]
    offset=0
    para_i=0
    for pi,para_txt in enumerate(paras_txt):
        if pi: para_txt=sep_para+para_txt
        para_txt_len=len(para_txt)
        para_str=para_txt.strip()
        para_str_len=len(para_str)
        para_str_noleft=para_txt.lstrip()
        para_str_noright=para_txt.rstrip()
        diff_left=para_txt_len - len(para_str_noleft)
        diff_right=para_txt_len - len(para_str_noright)
        
        offset1=offset+diff_left
        offset2=offset+para_txt_len-diff_right

        if para_str:
            assert para_str == txt[offset1:offset2]
            para_i+=1
            dx={
                'para_i':para_i,
                'para_str':para_str,
                'para_start_char':offset1,
                'para_end_char':offset2,
                # 'para_txt':para_txt,
            }
            ld.append(dx)
        offset+=para_txt_len
    return ld



# loading txt/strings
def to_txt(txt_or_fn):
    # load txt
    if type(txt_or_fn)==str and not '\n' in txt_or_fn and os.path.exists(txt_or_fn):
        with open(txt_or_fn,encoding='utf-8',errors='replace') as f:
            return f.read()
    return txt_or_fn



SIZELIM=53687091200

def get_db_scan(path=None):
    if not path or not os.path.exists(path): path=PATH_DB_SCAN
    return get_db(path=path)

def get_db_parse(path=None):
    if not path or not os.path.exists(path): path=PATH_DB_PARSE
    return get_db(path=path)

def get_db(path=None):
    if not path or not os.path.exists(path): path=PATH_DB
    return dc.Cache(path,size_limit=SIZELIM)


def get_linepart_str(lpdf_or_str):
    if type(lpdf_or_str)==str: return lpdf_or_str
    elif type(lpdf_or_str)==pd.DataFrame:
        lpdf=resetindex(lpdf_or_str)
        return ''.join(lpdf[(lpdf.word_ipa_i==1) & (lpdf.syll_i==1)].word_str)
    return ''

def get_linepart_df(lpdf_or_str,**kwargs):
    if type(lpdf_or_str)==pd.DataFrame: return resetindex(lpdf_or_str)
    elif type(lpdf_or_str)==str: return resetindex(get_scansion(lpdf_or_str,**kwargs))
    return pd.DataFrame()








sonnet="""How heavy do I journey on the way,
When what I seek, my weary travel's end,
Doth teach that ease and that repose to say
'Thus far the miles are measured from thy friend!'
The beast that bears me, tired with my woe,
Plods dully on, to bear that weight in me,
As if by some instinct the wretch did know
His rider loved not speed, being made from thee:
The bloody spur cannot provoke him on
That sometimes anger thrusts into his hide;
Which heavily he answers with a groan,
More sharp to me than spurring to his side;
For that same groan doth put this in my mind;
My grief lies onward and my joy behind."""
sonnet1="How heavy do I journey on the way,"




from ..imports import *
from .lexconvert import pipeThroughEspeak
from .lexconvert import convert as lexconvert
import nltk





Pyphen = None

# lang specific constants
PATH_LANG_DATA=os.path.join(PATH_DATA,'en')
CMU_DICT_FN=os.path.join(PATH_LANG_DATA,'cmudict.txt')
CMU_DICT={}
CACHE_DICT_FN=os.path.join(PATH_LANG_DATA,'tts-cache.txt')
CACHE_DICT_F=None
CACHE=defaultdict(list)
ORTH_CACHE=defaultdict(list)
CONTRACTIONS={"n't","'ve","'ll","'m","'d","'s","'m","'re"}
SPECIALD={}

def get_special_cases():
    global SPECIALD
    
    if not SPECIALD:

        ufn=os.path.join(PATH_LANG_DATA,'unstressed.txt')
        if os.path.exists(ufn):
            with open(ufn) as f:
                SPECIALD['unstressed']={w.strip().lower() for w in f.read().strip().split()}

        ufn=os.path.join(PATH_LANG_DATA,'maybestressed.txt')
        if os.path.exists(ufn):
            with open(ufn) as f:
                SPECIALD['maybestressed']={w.strip().lower() for w in f.read().strip().split()}

        import nltk
        #SPECIALD['functionwords']=set(nltk.corpus.stopwords.words('english'))
        SPECIALD['functionwords']=SPECIALD['unstressed']
        SPECIALD['functionwords']|=SPECIALD['maybestressed']

    return SPECIALD



"""
The only necessary functions
"""

def scan(line,incl_alt=True,**y):
    o=[
        {
            'word_i':word_i+1,
            **word_dx
        } for word_i,word in enumerate(tokenize(line))
        for word_dx in get(word,incl_alt=incl_alt)
    ]
    return pd.DataFrame(o)

def tokenize(txt,**y): 
    # return txt.split()
    # return tokenize_agnostic(txt,**y)
    return tokenize_nice(txt,**y)

def get_df(token,**kwargs):
    return pd.DataFrame(get(token,**kwargs))

def get(token,config={},toprint=False,incl_alt=True,cache_new=True,**kwargs):
    # not real?
    token_nice=token
    # token = zero_punc(token).lower()
    if not token: return []
    # get ipas
    cache = get_cache()
    # tokenl=token.lower()
    tokenl=to_token(token)
    if token in cache:
        ipas=cache[token]
    elif tokenl in cache:
        ipas=cache[tokenl]
    else:
        ipas=tts(tokenl)
        if ipas:
            cache[tokenl]=ipas
            if cache_new:
                write_to_cache(tokenl,ipas[0])
    ipas = ipas[:1] if not incl_alt else ipas
    sd=get_special_cases()
    if tokenl in sd['maybestressed']:
        if len(ipas)<2:
            ipa=ipas[0]
            ipax="'"+ipa if ipa[0].isalpha() else ipa[1:]
            ipas.append(ipax)
    elif tokenl in sd['unstressed']:
        ipa=ipas[0]
        ipax=ipa if ipa[0].isalpha() else ipa[1:]
        ipas=[ipax]

    is_funcword=tokenl in sd['functionwords']

    # get orths
    results = set()
    results_ld=[]
    for ipa_i,ipa in enumerate(ipas):
        num_sylls=ipa.count('.')+1
        sylls_ipa = ipa.split('.')
        sylls_text = syllabify_orth(token_nice,num_sylls=num_sylls)

        # maxlen=max([len(sylls_ipa), len(sylls_text)])
        
        # sylls_ipa+=['' for n in range(len(sylls_ipa) - maxlen)]
        # sylls_text+=['' for n in range(len(sylls_text) - maxlen)]

        is_funcword_now = int(is_funcword and num_sylls==1 and ipa[0]!="'")

        for si,(syll_ipa,syll_text) in enumerate(zip(sylls_ipa, sylls_text)):
            rdx={
                'word_ipa_i':ipa_i+1,
                'syll_i':si+1,
                # 'word_str':token_nice,
                'word_tok':tokenl,
                'word_ipa':ipa,
                'word_nsyll':num_sylls,
                'syll_ipa':syll_ipa,
                'syll_str':syll_text,
                'word_isfunc':is_funcword_now
            }
            results_ld.append(rdx)
    return results_ld


def get_cache(source_paths=[CMU_DICT_FN,CACHE_DICT_FN]):
# def get_cache(source_paths=[CMU_DICT_FN]):
    if not CACHE:
        for sfn in source_paths:
            if not os.path.exists(sfn): continue
            with open(sfn, encoding='utf-8') as f:
                for ln in f:
                    ln=ln.strip()
                    if not ln: continue
                    if ln.startswith('#'): continue
                    if not '\t' in ln: continue
                    word,ipa=[x.strip() for x in ln.strip().split('\t',1)]
                    if ipa not in set(CACHE[word]):
                        CACHE[word].append(ipa)
    return CACHE



def tts(token):
    espeak_ipa=espeak2ipa(token)
    if espeak_ipa is None: return []
    cmu=espeak2cmu(espeak_ipa)
    cmu_sylls = syllabify_cmu(cmu)
    ipa = cmusylls2ipa(cmu_sylls)
    return [ipa] if ipa else []




def write_to_cache(token,ipa):
    with open(CACHE_DICT_FN,'a+',encoding='utf-8') as of:
        of.write(f'{token}\t{ipa}\n')


def add_elisions(_ipa):
    """
    Add alternative pronunciations: those that have elided syllables
    """
    replace={}

    # -OWER
    # e.g. tower, hour, bower, etc
    replace['aʊ.ɛː']='aʊr'


    # -INOUS
    # e.g. ominous, etc
    replace['ə.nəs']='nəs'

    # -EROUS
    # e.g. ponderous, adventurous
    replace['ɛː.əs']='rəs'

    # -IA-
    # e.g. plutonian, indian, assyrian, idea, etc
    replace['iː.ə']='jə'
    # -IOUS
    # e.g. studious, tedious, etc
    #replace[u'iː.əs']=u'iːəs'


    # -IER
    # e.g. happier
    replace['iː.ɛː']='ɪr'

    # -ERING
    # e.g. scattering, wondering, watering
    replace['ɛː.ɪŋ']='rɪŋ'

    # -ERY
    # e.g. memory
    # QUESTIONABLE
    #replace[u'ɛː.iː']=u'riː'

    # -ENING
    # e.g. opening
    replace['ə.nɪŋ']='nɪŋ'

    # -ENER
    # e.g. gardener
    replace['ə.nɛː']='nɛː'

    # -EL- (-ELLER, -ELLING, -ELLY)
    # e.g. traveller, dangling, gravelly
    # QUESTIONABLE
    #replace[u'ə.l']=u'l'

    # -IRE-
    # e.g. fire, fiery, attire, hired
    replace['ɪ.ɛː']='ɪr'

    # -EL, -UAL
    # e.g. jewel
    replace['uː.əl']='uːl'

    # -EVN
    # e.g. heaven, seven
    replace['ɛ.vən']='ɛvn'

    # -IOUS, -EER
    # e.g. sincerest, dear, incommodiously
    # QUESTIONABLE
    #replace[u'.ʌ.']=u'ʌ.'
    replace['eɪ.ʌ']='eɪʌ'

    new=[_ipa]
    for k,v in list(replace.items()):
        if k in _ipa:
            new+=[_ipa.replace(k,v)]
    return new




import shutil
ERROR_SHOWN=False
ERROR_MSG="""
[cadence] A word unknown to the pronunciation dictionary ("{{WORD}}") cannot be syllabified because espeak text-to-speech software is not installed. It is highly recommend you install this software.

To install:

    * On Linux, type into the terminal:
        apt-get install espeak
    
    * On Mac:
        (1) Install homebrew if not already installed. Paste into Terminal app:
            /bin/bash -c "$(curl -fsSL https://raw.githubusercontent.com/Homebrew/install/HEAD/install.sh)"

        (2) Type into the Terminal app:
            brew install espeak
    
    * On Windows:
        Download and install from http://espeak.sourceforge.net/download.html.
"""

def espeak2ipa(token):
    global ERROR_SHOWN
    if not shutil.which('espeak'):
        eprint(ERROR_MSG.replace('{{WORD}}',token))
        ERROR_SHOWN=True
        return None
    
    token=''.join(x for x in token if x.isalpha())
    res = pipeThroughEspeak(token.encode('utf-8'))
    try:
        return res.decode('utf-8').strip()
    except Exception as e:
        return None

    # CMD=f'espeak -q -x {token}'
    # try:
    #     # res=subprocess.check_output(CMD, shell=True).strip()
    #     p = subprocess.Popen(command, stdout=subprocess.PIPE, stdin=infile)
    #  text = p.stdout.read()

    #     return res.decode("utf-8")
    # except (OSError,subprocess.CalledProcessError) as e:
    #     return None

def tts2ipa(token,TTS_ENGINE=None):
    if TTS_ENGINE=='espeak':
        return espeak2ipa(token)
    elif TTS_ENGINE=='openmary':
        return openmary2ipa(token)
    else:
        #raise Exception("No TTS engine specified. Please select 'espeak' or 'openmary' in config.txt.")
        return None

def espeak2cmu(tok):
    return lexconvert(str(tok),'espeak','cmu')

def ipa2cmu(tok):
    return lexconvert(str(tok),'unicode-ipa','cmu')

def cmu2ipa(tok):
    res=lexconvert(str(tok),'cmu','unicode-ipa')

    ## BUG FIXES
    if tok.endswith(' T') and not res.endswith('t'): res=res+'t'
    return res

def syllabify_cmu(cmu_token):
    cmu_token=cmu_token.replace(' 2','2').replace(' 1','1') # fix prim/sec stress markings for syllabify
    sylls = syllabify(English, cmu_token)
    return sylls



def cmusylls2ipa1(sylls):
    new_cmu=[]
    for syl in sylls:
        stress, onset, nucleus, coda = syl
        """if not stress:
            stress_str=""
        elif stress==1:
            stress_str="'"
            stress_str=str(stress)
        else:
            stress_str="`"
            stress_str=str(stress)
        """
        if stress:
            nucleus = [nucleus[0]+" "+str(stress)] + nucleus[1:]

        _newcmu = " ".join(onset+nucleus+coda).strip().replace("  "," ")
        new_cmu+=[_newcmu]
    new_cmu =" 0 ".join(new_cmu)
    #print new_cmu
    print(new_cmu)

    ## ipa
    ipa=cmu2ipa(new_cmu)
    print(ipa)
    #print ipa
    # clean
    ipa_sylls = ipa.split('.')
    for i,syl in enumerate(ipa_sylls):
        if "ˈ" in syl:
            syl="'"+syl.replace("ˈ","")
        if "ˌ" in syl:
            syl="`"+syl.replace("ˌ","")
        ipa_sylls[i]=syl
    ipa=".".join(ipa_sylls)
    print(ipa)
    return ipa

def cmusylls2ipa(sylls):
    new_cmu=[]
    new_ipa=[]
    for syl in sylls:
        stress, onset, nucleus, coda = syl
        if stress:
            nucleus = [nucleus[0]+" "+str(stress)] + nucleus[1:]

        _newcmu = " ".join(onset+nucleus+coda).strip().replace("  "," ")
        _newipa=cmu2ipa(_newcmu)
        new_ipa+=[_newipa]
        new_cmu+=[_newcmu]
    new_cmu =" 0 ".join(new_cmu)
    #print new_cmu
    #print(new_cmu)

    ## ipa
    ipa=cmu2ipa(new_cmu)
    #print(ipa)
    #print ipa
    # clean
    #ipa_sylls = ipa.split('.')
    ipa_sylls = new_ipa
    for i,syl in enumerate(ipa_sylls):
        if "ˈ" in syl:
            syl="'"+syl.replace("ˈ","")
        if "ˌ" in syl:
            syl="`"+syl.replace("ˌ","")
        ipa_sylls[i]=syl
    ipa=".".join(ipa_sylls)
    #print(ipa)
    return ipa


def syllabify_orth_with_hyphenate(token,num_sylls=None):
    from hyphenate import hyphenate_word
    return hyphenate_word(token)
    return l
nltk_ssp=None
def syllabify_orth_with_nltk(token,num_sylls=None):
    global nltk_ssp
    if not nltk_ssp:
        from nltk.tokenize import SyllableTokenizer
        nltk_ssp = SyllableTokenizer()
    tokenl=token.lower()
    l = nltk_ssp.tokenize(tokenl)
    if tokenl!=token:
        o=[]
        i=0
        for x in l:
            xlen=len(x)
            o+=[token[i:i+xlen]]
            i+=xlen
        l=o
    return l

def syllabify_orth_with_pyphen(token,num_sylls=None):
    global Pyphen
    if not Pyphen:
        import pyphen
        Pyphen=pyphen.Pyphen(lang='en_US')
    sylls = Pyphen.inserted(token,hyphen='||||').split('||||')
    return sylls

# def syllabify_orth(token,num_sylls=None, func=syllabify_orth_with_pyphen):
def syllabify_orth(token,num_sylls=None, func=syllabify_orth_with_nltk):
    key=(token,num_sylls)
    if not key in ORTH_CACHE:
        pref,tok,suf = split_punct(token)
        l=func(tok,num_sylls=num_sylls) if num_sylls>1 else [tok]
        while len(l)<num_sylls:
            lastsyll=l[-1]
            lastsyll_len_half=len(lastsyll)//2
            lastsyll_a,lastsyll_b=lastsyll[:lastsyll_len_half],lastsyll[lastsyll_len_half:]
            l=[sx for sx in l[:-1]] + [lastsyll_a, lastsyll_b]
        while len(l)>num_sylls:
            l2=l[:-1]
            l2[-1]+=l[-1]
            l=l2

        l[0]=pref+l[0]
        l[-1]+=suf

        ORTH_CACHE[key]=l
    return ORTH_CACHE[key]


















# This is the P2TK automated syllabifier. Given a string of phonemes,
# it automatically divides the phonemes into syllables.
#
# By Joshua Tauberer, based on code originally written by Charles Yang.
#
# The syllabifier requires a language configuration which specifies
# the set of phonemes which are consonants and vowels (syllable nuclei),
# as well as the set of permissible onsets.
#
# Then call syllabify with a language configuration object and a word
# represented as a string (or list) of phonemes.
#
# Returned is a data structure representing the syllabification.
# What you get is a list of syllables. Each syllable is a tuple
# of (stress, onset, nucleus, coda). stress is None or an integer stress
# level attached to the nucleus phoneme on input. onset, nucleus,
# and coda are lists of phonemes.
#
# Example:
#
# import syllabifier
# language = syllabifier.English # or: syllabifier.loadLanguage("english.cfg")
# syllables = syllabifier.syllabify(language, "AO2 R G AH0 N AH0 Z EY1 SH AH0 N Z")
#
# The syllables variable then holds the following:
# [ (2, [],     ['AO'], ['R']),
#   (0, ['G'],  ['AH'], []),
#   (0, ['N'],  ['AH'], []),
#   (1, ['Z'],  ['EY'], []),
#   (0, ['SH'], ['AH'], ['N', 'Z'])]
#
# You could process that result with this type of loop:
#
# for stress, onset, nucleus, coda in syllables :
#   print " ".join(onset), " ".join(nucleus), " ".join(coda)
#
# You can also pass the result to stringify to get a nice printable
# representation of the syllables, with periods separating syllables:
#
# print syllabify.stringify(syllables)
#
#########################################################################

English = {
    'consonants': ['B', 'CH', 'D', 'DH', 'F', 'G', 'HH', 'JH', 'K', 'L', 'M', 'N', 
    'NG', 'P', 'R', 'S', 'SH', 'T', 'TH', 'V', 'W', 'Y', 'Z', 'ZH'],
    'vowels': [ 'AA', 'AE', 'AH', 'AO', 'AW', 'AY', 'EH', 'ER', 'EY', 'IH', 'IY', 'OW', 'OY', 'UH', 'UW'],
    'onsets': ['P', 'T', 'K', 'B', 'D', 'G', 'F', 'V', 'TH', 'DH', 'S', 'Z', 'SH', 'CH', 'JH', 'M',
    'N', 'R', 'L', 'HH', 'W', 'Y', 'P R', 'T R', 'K R', 'B R', 'D R', 'G R', 'F R',
    'TH R', 'SH R', 'P L', 'K L', 'B L', 'G L', 'F L', 'S L', 'T W', 'K W', 'D W', 
    'S W', 'S P', 'S T', 'S K', 'S F', 'S M', 'S N', 'G W', 'SH W', 'S P R', 'S P L',
    'S T R', 'S K R', 'S K W', 'S K L', 'TH W', 'ZH', 'P Y', 'K Y', 'B Y', 'F Y', 
    'HH Y', 'V Y', 'TH Y', 'M Y', 'S P Y', 'S K Y', 'G Y', 'HH W', '']
    }

def loadLanguage(filename) :
    '''This function loads up a language configuration file and returns
    the configuration to be passed to the syllabify function.'''

    L = { "consonants" : [], "vowels" : [], "onsets" : [] }

    f = open(filename, "r")
    section = None
    for line in f :
        line = line.strip()
        if line in ("[consonants]", "[vowels]", "[onsets]") :
            section = line[1:-1]
        elif section == None :
            raise ValueError("File must start with a section header such as [consonants].")
        elif not section in L :
            raise ValueError("Invalid section: " + section)
        else :
            L[section].append(line)

    for section in "consonants", "vowels", "onsets" :
        if len(L[section]) == 0 :
            raise ValueError("File does not contain any consonants, vowels, or onsets.")

    return L

def syllabify(language, word) :
    '''Syllabifies the word, given a language configuration loaded with loadLanguage.
       word is either a string of phonemes from the CMU pronouncing dictionary set
       (with optional stress numbers after vowels), or a Python list of phonemes,
       e.g. "B AE1 T" or ["B", "AE1", "T"]'''

    if type(word) == str :
        word = word.split()

    syllables = [] # This is the returned data structure.

    internuclei = [] # This maintains a list of phonemes between nuclei.

    for phoneme in word :

        phoneme = phoneme.strip()
        if phoneme == "" :
            continue
        stress = None
        if phoneme[-1].isdigit() :
            stress = int(phoneme[-1])
            phoneme = phoneme[0:-1]

        if phoneme in language["vowels"] :
            # Split the consonants seen since the last nucleus into coda and onset.

            coda = None
            onset = None

            # If there is a period in the input, split there.
            if "." in internuclei :
                period = internuclei.index(".")
                coda = internuclei[:period]
                onset = internuclei[period+1:]

            else :
                # Make the largest onset we can. The 'split' variable marks the break point.
                for split in range(0, len(internuclei)+1) :
                    coda = internuclei[:split]
                    onset = internuclei[split:]

                    # If we are looking at a valid onset, or if we're at the start of the word
                    # (in which case an invalid onset is better than a coda that doesn't follow
                    # a nucleus), or if we've gone through all of the onsets and we didn't find
                    # any that are valid, then split the nonvowels we've seen at this location.
                    if " ".join(onset) in language["onsets"] \
                       or len(syllables) == 0 \
                       or len(onset) == 0 :
                       break

            # Tack the coda onto the coda of the last syllable. Can't do it if this
            # is the first syllable.
            if len(syllables) > 0 :
                syllables[-1][3].extend(coda)

            # Make a new syllable out of the onset and nucleus.
            syllables.append( (stress, onset, [phoneme], []) )

            # At this point we've processed the internuclei list.
            internuclei = []

        elif not phoneme in language["consonants"] and phoneme != "." :
            raise ValueError("Invalid phoneme: " + phoneme)

        else : # a consonant
            internuclei.append(phoneme)

    # Done looping through phonemes. We may have consonants left at the end.
    # We may have even not found a nucleus.
    if len(internuclei) > 0 :
        if len(syllables) == 0 :
            syllables.append( (None, internuclei, [], []) )
        else :
            syllables[-1][3].extend(internuclei)

    return syllables

def stringify(syllables) :
    '''This function takes a syllabification returned by syllabify and
       turns it into a string, with phonemes spearated by spaces and
       syllables spearated by periods.'''
    ret = []
    for syl in syllables :
        stress, onset, nucleus, coda = syl
        if stress != None and len(nucleus) != 0 :
            nucleus[0] += str(stress)
        ret.append(" ".join(onset + nucleus + coda))
    return " . ".join(ret)


# If this module was run directly, syllabify the words on standard input
# into standard output. Hashed lines are printed back untouched.
if __name__ == "__main__" :
    import sys
    if len(sys.argv) != 2 :
        print("Usage: python syllabifier.py english.cfg < textfile.txt > outfile.txt")
    else :
        L = loadLanguage(sys.argv[1])
        for line in sys.stdin :
            if line[0] == "#" :
                sys.stdout.write(line)
                continue
            line = line.strip()
            s = stringify(syllabify(L, line))
            sys.stdout.write(s + "\n")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

VERSION = (0, 0, 1)
__version__ = '.'.join(map(str, VERSION))

__all__ = [
        'format',
        'fullwidth',
        'styles',
        'FACE_PLAIN',
        'FACE_BOLD',
        'FACE_ITAL',
        'STYLE_SANS',
        'STYLE_SERIF',
        'STYLE_MONO',
        'STYLE_SCRIPT',
        'STYLE_DOUBLE',
        'STYLE_FRAKTUR',
        'STYLE_SMALLCAPS',
        ]

FACE_PLAIN = 0x00
FACE_BOLD = 0x01
FACE_ITAL = 0x02

STYLE_SANS = 0x0100
STYLE_SERIF = 0x0200
STYLE_MONO = 0x0400
STYLE_SCRIPT = 0x0800
STYLE_DOUBLE = 0x1000
STYLE_FRAKTUR = 0x2000
STYLE_SMALLCAPS = 0x4000

styles = {
        "sans" : STYLE_SANS,
        "serif" : STYLE_SERIF,
        "mono" : STYLE_MONO,
        "script" : STYLE_SCRIPT,
        "double" : STYLE_DOUBLE,
        "fraktur" : STYLE_FRAKTUR,
        "smallcaps" : STYLE_SMALLCAPS,
        }

valid_combos = set((
        STYLE_MONO,
        STYLE_SERIF|FACE_ITAL,
        STYLE_SERIF|FACE_BOLD,
        STYLE_SERIF|FACE_ITAL|FACE_BOLD,
        STYLE_SANS,
        STYLE_SANS|FACE_ITAL,
        STYLE_SANS|FACE_BOLD,
        STYLE_SANS|FACE_ITAL|FACE_BOLD,
        STYLE_SCRIPT,
        STYLE_SCRIPT|FACE_BOLD,
        STYLE_DOUBLE,
        STYLE_FRAKTUR,
        STYLE_FRAKTUR|FACE_BOLD,
        STYLE_SMALLCAPS,
        ))

charmap = {}
for c in valid_combos:
    charmap[c] = {}

## We can only convert a subset of text, specifically A-Za-z0-9.
for i in range(0, 26):
    cap = chr(i + ord('A'))
    sml = chr(i + ord('a'))

    charmap[STYLE_MONO][cap] = chr(i + 0x1d670)
    charmap[STYLE_MONO][sml] = chr(i + 0x1d68a)

    # Use IT/it/BD/bd as shortcuts to the italicized/bolded letters
    # so we can then map the opposite style applied over them onto
    # bold+italic
    IT = charmap[STYLE_SERIF|FACE_ITAL][cap] = chr(i + 0x1d434)
    it = charmap[STYLE_SERIF|FACE_ITAL][sml] = chr(i + 0x1d44e)
    BD = charmap[STYLE_SERIF|FACE_BOLD][cap] = chr(i + 0x1d400)
    bd = charmap[STYLE_SERIF|FACE_BOLD][sml] = chr(i + 0x1d41a)

    charmap[STYLE_SERIF|FACE_BOLD][IT] = \
    charmap[STYLE_SERIF|FACE_ITAL][BD] = \
    charmap[STYLE_SERIF|FACE_BOLD|FACE_ITAL][cap] = chr(i + 0x1d468)
    charmap[STYLE_SERIF|FACE_BOLD][it] = \
    charmap[STYLE_SERIF|FACE_ITAL][bd] = \
    charmap[STYLE_SERIF|FACE_BOLD|FACE_ITAL][sml] = chr(i + 0x1d482)

    IT = charmap[STYLE_SANS|FACE_ITAL][cap] = chr(i + 0x1d608)
    it = charmap[STYLE_SANS|FACE_ITAL][sml] = chr(i + 0x1d622)
    BD = charmap[STYLE_SANS|FACE_BOLD][cap] = chr(i + 0x1d5d4)
    bd = charmap[STYLE_SANS|FACE_BOLD][sml] = chr(i + 0x1d5ee)

    charmap[STYLE_SANS][cap] = chr(i + 0x1d5a0)
    charmap[STYLE_SANS][sml] = chr(i + 0x1d5ba)

    charmap[STYLE_SANS|FACE_BOLD][IT] = \
    charmap[STYLE_SANS|FACE_ITAL][BD] = \
    charmap[STYLE_SANS|FACE_BOLD|FACE_ITAL][cap] = chr(i + 0x1d63c)
    charmap[STYLE_SANS|FACE_BOLD][it] = \
    charmap[STYLE_SANS|FACE_ITAL][bd] = \
    charmap[STYLE_SANS|FACE_BOLD|FACE_ITAL][sml] = chr(i + 0x1d656)

    charmap[STYLE_SCRIPT][cap] = chr(i + 0x1d49c)
    charmap[STYLE_SCRIPT][sml] = chr(i + 0x1d4b6)
    charmap[STYLE_SCRIPT|FACE_BOLD][cap] = chr(i + 0x1d4d0)
    charmap[STYLE_SCRIPT|FACE_BOLD][sml] = chr(i + 0x1d4ea)

    charmap[STYLE_DOUBLE][cap] = chr(i + 0x1d538)
    charmap[STYLE_DOUBLE][sml] = chr(i + 0x1d552)

    charmap[STYLE_FRAKTUR][cap] = chr(i + 0x1d504)
    charmap[STYLE_FRAKTUR][sml] = chr(i + 0x1d51e)
    charmap[STYLE_FRAKTUR|FACE_BOLD][cap] = chr(i + 0x1d56c)
    charmap[STYLE_FRAKTUR|FACE_BOLD][sml] = chr(i + 0x1d586)

    charmap[STYLE_SMALLCAPS][cap] = cap

for i in range(0, 10):
    d = chr(i + ord('0'))
    charmap[STYLE_MONO][d] = chr(i + 0x1d7f6)
    charmap[STYLE_SERIF|FACE_BOLD][d] = chr(i + 0x1d7ce)
    charmap[STYLE_SANS|FACE_BOLD][d] = chr(i + 0x1d7ec)
    charmap[STYLE_DOUBLE][d] = chr(i + 0x1d7d8)
    charmap[STYLE_SANS][d] = chr(i + 0x1d7e2)

# Not all glyphs are in nice codepoint sequence, patch the gaps here
charmap[STYLE_SERIF|FACE_ITAL]["h"] = chr(0x210e)
charmap[STYLE_SCRIPT]["B"] = chr(0x212c)
charmap[STYLE_SCRIPT]["E"] = chr(0x2130)
charmap[STYLE_SCRIPT]["F"] = chr(0x2131)
charmap[STYLE_SCRIPT]["H"] = chr(0x210b)
charmap[STYLE_SCRIPT]["I"] = chr(0x2110)
charmap[STYLE_SCRIPT]["L"] = chr(0x2112)
charmap[STYLE_SCRIPT]["M"] = chr(0x2133)
charmap[STYLE_SCRIPT]["R"] = chr(0x2113)
charmap[STYLE_SCRIPT]["e"] = chr(0x212f)
charmap[STYLE_SCRIPT]["g"] = chr(0x210a)
charmap[STYLE_SCRIPT]["o"] = chr(0x2134)
charmap[STYLE_DOUBLE]["C"] = chr(0x2102)
charmap[STYLE_DOUBLE]["H"] = chr(0x210d)
charmap[STYLE_DOUBLE]["N"] = chr(0x2115)
charmap[STYLE_DOUBLE]["P"] = chr(0x2119)
charmap[STYLE_DOUBLE]["Q"] = chr(0x211a)
charmap[STYLE_DOUBLE]["R"] = chr(0x211d)
charmap[STYLE_DOUBLE]["Z"] = chr(0x2124)
charmap[STYLE_FRAKTUR]["C"] = chr(0x212d)
charmap[STYLE_FRAKTUR]["H"] = chr(0x210c)
charmap[STYLE_FRAKTUR]["I"] = chr(0x2111)
charmap[STYLE_FRAKTUR]["R"] = chr(0x211c)
charmap[STYLE_FRAKTUR]["Z"] = chr(0x2128)

# And smallcaps are all over the place
charmap[STYLE_SMALLCAPS]["a"] = chr(0x1d00)
charmap[STYLE_SMALLCAPS]["b"] = chr(0x0299)
charmap[STYLE_SMALLCAPS]["c"] = chr(0x1d04)
charmap[STYLE_SMALLCAPS]["d"] = chr(0x1d05)
charmap[STYLE_SMALLCAPS]["e"] = chr(0x1d07)
charmap[STYLE_SMALLCAPS]["f"] = chr(0xa730)
charmap[STYLE_SMALLCAPS]["g"] = chr(0x0262)
charmap[STYLE_SMALLCAPS]["h"] = chr(0x029c)
charmap[STYLE_SMALLCAPS]["i"] = chr(0x026a)
charmap[STYLE_SMALLCAPS]["j"] = chr(0x1d0a)
charmap[STYLE_SMALLCAPS]["k"] = chr(0x1d0b)
charmap[STYLE_SMALLCAPS]["l"] = chr(0x029f)
charmap[STYLE_SMALLCAPS]["m"] = chr(0x1d0d)
charmap[STYLE_SMALLCAPS]["n"] = chr(0x0274)
charmap[STYLE_SMALLCAPS]["o"] = chr(0x1d0f)
charmap[STYLE_SMALLCAPS]["p"] = chr(0x1d18)
charmap[STYLE_SMALLCAPS]["q"] = "Q" # Can't find this one anywhere
charmap[STYLE_SMALLCAPS]["r"] = chr(0x0280)
charmap[STYLE_SMALLCAPS]["s"] = chr(0xa731)
charmap[STYLE_SMALLCAPS]["t"] = chr(0x1d1b)
charmap[STYLE_SMALLCAPS]["u"] = chr(0x1d1c)
charmap[STYLE_SMALLCAPS]["v"] = chr(0x1d20)
charmap[STYLE_SMALLCAPS]["w"] = chr(0x1d21)
charmap[STYLE_SMALLCAPS]["x"] = "x"
charmap[STYLE_SMALLCAPS]["y"] = chr(0x028f)
charmap[STYLE_SMALLCAPS]["z"] = chr(0x1d22)


def format(fmt, st):
    """\
    Return the formatted verison of the given string.
    """
    ret = ""
    if not st: return ret
    if fmt not in valid_combos:
        return st
    cm = charmap[fmt]
    for c in st:
        ret += cm.get(c, c)
    return ret

def fullwidth(st):
    """\
    Return the fullwidth version of the given string.
    """
    ret = ""
    if not st: return ret
    for c in st:
        i = ord(c)
        if c == " ":
            ret += chr(0x3000)
        elif 0x21 <= i <= 0x7f:
            ret += chr(i - 0x21 + 0xff01)
        else:
            ret += c
    return ret

# Self-test routine
if __name__ == '__main__':

    style_names = { styles[k]:k for k in styles.keys() }
    combo_names = {}
    longest_name = 0
    for c in valid_combos:
        sty = c & ~0xFF
        cn = style_names[sty]
        if c & FACE_BOLD:
            cn += " bold"
        if c & FACE_ITAL:
            cn += " ital"
        combo_names[c] = cn
        longest_name = max(longest_name, len(cn))

    name_fmt = "{{:>{0}s}}: ".format(longest_name)

    for c in valid_combos:
        print(name_fmt.format(combo_names[c]), end="")
        print(format(c,
            "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"))

    print(format(STYLE_SANS|FACE_ITAL|FACE_BOLD,
        "pre-combined bold and italic"))
    print(format(STYLE_SANS|FACE_ITAL,
        format(STYLE_SANS|FACE_BOLD, "Chained bold and italic")
        ))

#
# Editor modelines - http://www.wireshark.org/tools/modelines.html
#
# Local variables:
# c-basic-offset: 4
# tab-width: 4
# indent-tabs-mode: nil
# End:
#
# vi:set shiftwidth=4 tabstop=4 expandtab:
# indentSize=4:tabSize=4:noTabs=true:


#!/usr/bin/env python # -*- coding: utf-8 -*-
#8/31/15: 1200-1600
#9/06/15: 1330-1945
#9/08/15: 1130-1430
#9/15 3:30-6:30 <- taken care of?
#11/25 eh, like 30 mins

import os
from collections import defaultdict
import pickle as pkl
import numpy as np
#import pandas as pd
#import matplotlib.pyplot as plt
import codecs
import nltk
from nltk import compat
from nltk.tree import Tree
import nltk.data
from .deptree import DependencyTree#, DependencyTreeParser
import re,time


DIR_ROOT = '.'
sylcmu = {}

INFO_DO_NOT_STORE = ['contour','sent','nsyll','nseg','seg','nstress','word']



#***********************************************************************
# Metrical Tree class
class MetricalTree(DependencyTree):
    """"""

    _unstressedWords = ('it',)
    _unstressedTags  = ('CC', 'PRP$', 'TO', 'UH', 'DT')
    _unstressedDeps  = ('det', 'expl', 'cc', 'mark')
    _ambiguousWords = ('this', 'that', 'these', 'those')
    _ambiguousTags  = ('MD', 'IN', 'PRP', 'WP$', 'PDT', 'WDT', 'WP', 'WRB')
    _ambiguousDeps  = ('cop', 'neg', 'aux', 'auxpass')
    _stressedWords = tuple()

    #=====================================================================
    # Initialize
    def __init__(self, node, children, dep=None, lstress=0, pstress=np.nan, stress=np.nan):
        """"""

        self._lstress = lstress
        self._pstress = pstress
        self._stress = stress
        super(MetricalTree, self).__init__(node, children, dep)
        self.set_label()
        if self._preterm:
            if self[0].lower() in sylcmu:
                syll_info = sylcmu[self[0].lower()]
                self._seg = syll_info[0]
                self._nsyll = len(syll_info[1])
                self._nstress = len([x for x in syll_info[1] if x[1] in ('P', 'S')])
            else:
                self._seg = None
                self._nsyll = np.nan
                self._nstress = np.nan

    #=====================================================================
    # Get the lexical stress of the node
    def lstress(self):
        """"""

        return self._lstress

    #=====================================================================
    # Get the phrasal stress of the node
    def pstress(self):
        """"""

        return self._pstress

    #=====================================================================
    # Get the stress of the node
    def stress(self):
        """"""

        return self._stress

    #=====================================================================
    # Get the segments
    def seg(self):
        """"""

        return self._seg if self._seg is not None else []

    #=====================================================================
    # Get the number of segments
    def nseg(self):
        """"""

        return len(self._seg) if self._seg is not None else np.nan

    #=====================================================================
    # Get the number of syllables
    def nsyll(self):
        """"""

        return self._nsyll

    #=====================================================================
    # Get the number of stresses
    def nstress(self):
        """"""

        return self._nstress

    #=====================================================================
    # Get the lexical stress of the leaf nodes
    def lstresses(self, leaves=True):
        """"""

        for preterminal in self.preterminals(leaves=True):
            if leaves:
                yield (preterminal._lstress, preterminal[0])
            else:
                yield preterminal._lstress

    #=====================================================================
    # Get the phrasal stress of the leaf nodes
    def pstresses(self, leaves=True):
        """"""

        for preterminal in self.preterminals(leaves=True):
            if leaves:
                yield (preterminal._pstress, preterminal[0])
            else:
                yield preterminal._pstress

    #=====================================================================
    # Get the lexical stress of the leaf nodes
    def stresses(self, leaves=True, arto=False):
        """"""

        for preterminal in self.preterminals(leaves=True):
            if leaves:
                if arto:
                    if preterminal._stress is None:
                        yield (None, preterminal[0])
                    elif preterminal._lstress == -1:
                        yield (0, preterminal[0])
                    else:
                        yield (-(preterminal._stress-1), preterminal[0])
                else:
                    yield (preterminal._stress, preterminal[0])
            else:
                if arto:
                    if preterminal._stress is None:
                        yield None
                    elif preterminal._lstress == -1:
                        yield 0
                    else:
                        yield -(preterminal._stress-1)
                else:
                    yield preterminal._stress

    #=====================================================================
    # Get the number of syllables of the leaf nodes
    def nsylls(self, leaves=True):
        """"""

        for preterminal in self.preterminals(leaves=True):
            if leaves:
                yield (preterminal._nsyll, preterminal[0])
            else:
                yield preterminal._nsyll

    #=====================================================================
    # Set the lexical stress of the node
    def set_lstress(self):
        """"""

        if self._preterm:
            if self[0].lower() in super(MetricalTree, self)._contractables:
                self._lstress = np.nan
            elif self._cat in super(MetricalTree, self)._punctTags:
                self._lstress = np.nan

            elif self[0].lower() in MetricalTree._unstressedWords:
                self._lstress = -1
            elif self[0].lower() in MetricalTree._ambiguousWords:
                self._lstress = -.5
            elif self[0].lower() in MetricalTree._stressedWords:
                self._lstress = 0

            elif self._cat in MetricalTree._unstressedTags:
                self._lstress = -1
            elif self._cat in MetricalTree._ambiguousTags:
                self._lstress = -.5

            elif self._dep in MetricalTree._unstressedDeps:
                self._lstress = -1
            elif self._dep in MetricalTree._ambiguousDeps:
                self._lstress = -.5

            else:
                self._lstress = 0

            if self[0].lower() == 'that' and (self._cat == 'DT' or self._dep == 'det'):
                self._lstress = -.5
        else:
            for child in self:
                child.set_lstress()
        self.set_label()

    #=====================================================================
    # Set the phrasal stress of the tree
    def set_pstress(self):
        """"""

        # Basis
        if self._preterm:
            try: assert self._lstress != -.5
            except: raise ValueError('The tree must be disambiguated before assigning phrasal stress')
            self._pstress = self._lstress
        else:
            # Recurse
            for child in self:
                child.set_pstress()
            assigned = False
            # Noun compounds (look for sequences of N*)
            if self._cat == 'NP':
                skipIdx = None
                i = len(self)
                for child in self[::-1]:
                    i -= 1
                    if child._cat.startswith('NN'):
                        if not assigned and skipIdx is None:
                            skipIdx = i
                            child._pstress = -1
                            child.set_label()
                        elif not assigned:
                            child._pstress = 0
                            child.set_label()
                            assigned = True
                        else:
                            child._pstress = -1
                            child.set_label()
                    elif assigned:
                        child._pstress = -1
                        child.set_label()
                    else:
                        if not assigned and skipIdx is not None:
                            self[skipIdx]._pstress = 0
                            self[skipIdx].set_label()
                            assigned = True
                            child._pstress = -1
                            child.set_label()
                        else:
                            break
                if not assigned and skipIdx is not None:
                    self[skipIdx]._pstress = 0
                    self[skipIdx].set_label()
                    assigned = True
            # Everything else
            if not assigned:
                for child in self[::-1]:
                    if not assigned and child._pstress == 0:
                        assigned = True
                    elif not np.isnan(child._pstress):
                        child._pstress = -1
                        child.set_label()
            if not assigned:
                self._pstress = -1
            else:
                self._pstress = 0
        self.set_label()

    #=====================================================================
    # Set the total of the tree
    def set_stress(self, stress=0):
        """"""

        self._stress = self._lstress + self._pstress + stress
        if not self._preterm:
            for child in self:
                child.set_stress(self._stress)
        self.set_label()

    #=====================================================================
    # Reset the label of the node (cat < dep < lstress < pstress < stress
    def set_label(self):
        """"""

        if self._stress is not np.nan:
            self._label = '%s/%s' % (self._cat, self._stress)
        elif self._pstress is not np.nan:
            self._label = '%s/%s' % (self._cat, self._pstress)
        elif self._lstress is not np.nan:
            self._label = '%s/%s' % (self._cat, self._lstress)
        elif self._dep is not None:
            self._label = '%s/%s' % (self._cat, self._dep)
        else:
            self._label = '%s' % self._cat

    #=====================================================================
    # Convert between different subtypes of Metrical Trees
    @classmethod
    def convert(cls, tree):
        """
        Convert a tree between different subtypes of Tree.  ``cls`` determines
        which class will be used to encode the new tree.

        :type tree: Tree
        :param tree: The tree that should be converted.
        :return: The new Tree.
        """

        if isinstance(tree, Tree):
            children = [cls.convert(child) for child in tree]
            if isinstance(tree, MetricalTree):
                return cls(tree._cat, children, tree._dep, tree._lstress)
            elif isinstance(tree, DependencyTree):
                return cls(tree._cat, children, tree._dep)
            else:
                return cls(tree._label, children)
        else:
            return tree

    #=====================================================================
    # Approximate the number of ambiguous parses
    def ambiguity(self, stress_polysyll=False):
        """"""

        nambig = 0
        for preterminal in self.preterminals():
            if preterminal.lstress() == -.5:
                if not stress_polysyll or (preterminal.nsyll() == 1):
                    nambig += 1
        return nambig

    #=====================================================================
    # Generate all possible trees
    # Syll=True sets all polysyllabic words to stressed
    def ambiguate(self, stress_polysyll=False):
        """"""

        if self._preterm:
            if self._lstress != -.5:
                return [self.copy()]
            else:
                alts = []
                if not stress_polysyll or self._nsyll == 1:
                    self._lstress = -1
                    alts.append(self.copy())
                self._lstress = 0
                alts.append(self.copy())
                self._lstress = -.5
                return alts
        else:
            alts = [[]]
            for child in self:
                child_alts = child.disambiguate(syll)
                for i in range(len(alts)):
                    alt = alts.pop(0)
                    for child_alt in child_alts:
                        alts.append(alt + [child_alt])
            return [MetricalTree(self._cat, alt, self._dep) for alt in alts]

    #=====================================================================
    # Disambiguate a tree with the maximal stressed pattern
    def max_stress_disambiguate(self):
        """"""

        if self._preterm:
            if self._lstress != -.5:
                return [self.copy()]
            else:
                alts = []
                self._lstress = 0
                alts.append(self.copy())
                self._lstress = -.5
                return alts
        else:
            alts = [[]]
            for child in self:
                child_alts = child.max_stress_disambiguate()
                for i in range(len(alts)):
                    alt = alts.pop(0)
                    for child_alt in child_alts:
                        alts.append(alt + [child_alt])
            return [MetricalTree(self._cat, alt, self._dep) for alt in alts]

    #=====================================================================
    # Disambiguate a tree with the minimal stressed pattern
    def min_stress_disambiguate(self, stress_polysyll=False):
        """"""

        if self._preterm:
            if self._lstress != -.5:
                return [self.copy()]
            else:
                alts = []
                if not stress_polysyll or self._nsyll == 1:
                    self._lstress = -1
                else:
                    self._lstress = 0

                alts.append(self.copy())
                self._lstress = -.5
                return alts
        else:
            alts = [[]]
            for child in self:
                child_alts = child.min_stress_disambiguate(stress_polysyll)
                for i in range(len(alts)):
                    alt = alts.pop(0)
                    for child_alt in child_alts:
                        alts.append(alt + [child_alt])
            return [MetricalTree(self._cat, alt, self._dep) for alt in alts]

    #=====================================================================
    # Copy the tree
    def copy(self, deep=False):
        """"""

        if not deep:
            return type(self)(self._cat, self, dep=self._dep, lstress=self._lstress)
        else:
            return type(self).convert(self)



from metricaltree import *

parser = DependencyTreeParser(model_path='Stanford Library/stanford-parser-full-%s/edu/stanford/nlp/models/lexparser/englishPCFG.ser.gz' % DATE)
parser = MetricalTreeParser(parser)

sents = ['hello beautiful world!', 'how are you?']

t = list(parser.lex_raw_parse_sents(sents))[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Natural Language Toolkit: Updated interface to the Stanford Parser
#
# Copyright (C) 2015 Tim Dozat
# Author: Tim Dozat <tdozat@stanford.edu>
# Author of the Stanford Parser nltk code: Steven Xu <xxu@student.unimelb.edu.au>
#
# For license information, see LICENSE.TXT




import tempfile
import os
import re
from subprocess import PIPE

import nltk
import nltk.data
from nltk import compat
from nltk import Tree
from nltk.internals import find_jar, find_jar_iter, config_java, java, _java_options

from nltk.parse.api import ParserI

_stanford_url = 'http://nlp.stanford.edu/software/lex-parser.shtml'

#***********************************************************************
# Dependency-augmented syntactic tree class
class DependencyTree(Tree):
    """"""

    _contractables = ("m", "s", "ll", "d", "nt", "re", "ve", "'m", "'s", "'ll", "'d", "n't", "'re", "'ve")
    _punctTags = ('.', ',', ':')

    #=====================================================================
    # Initialize
    def __init__(self, node, children=None, dep=None):
        """"""

        self._cat = node
        self._dep = dep
        self._preterm = False
        self._label = None
        super(DependencyTree, self).__init__(node, children)
        if len(self) == 1 and isinstance(self[0], str):
            self._preterm = True
        self.set_label()

    #=====================================================================
    # Get the preterminal value of the node
    def preterminal(self):
        """"""

        return self._preterm

    #=====================================================================
    # Get the categorial value of the node
    def category(self):
        """"""

        return self._cat

    #=====================================================================
    # Get the dependency label of the node
    def dependency(self):
        """"""

        return self._dep

    #=====================================================================
    # Get the dependency labels of the leaf nodes
    def preterminals(self, leaves=True):
        """"""

        if self._preterm:
            if leaves:
                yield self
            else:
                yield self._label
        else:
            for child in self:
                for preterminal in child.preterminals(leaves=leaves):
                    yield preterminal

    #=====================================================================
    # Get the category labels of the leaf nodes
    def categories(self, leaves=True):
        """"""

        for preterminal in self.preterminals(leaves=True):
            if leaves:
                yield (preterminal._cat, preterminal[0])
            else:
                yield preterminal._cat

    #=====================================================================
    # Get the dependency labels of the leaf nodes
    def dependencies(self, leaves=True):
        """"""

        for preterminal in self.preterminals(leaves=True):
            if leaves:
                yield (preterminal._dep, preterminal[0])
            else:
                yield preterminal._dep

    #=====================================================================
    # Reset the label of the node
    def set_label(self):
        """"""

        if self._dep is None:
            self._label = self._cat
        else:
            self._label = '%s/%s' % (self._cat, self._dep)

    #=====================================================================
    # Set the category of the node
    def set_category(self, cat):
        """"""

        self._cat = cat
        self.set_label()

    #=====================================================================
    # Set the dependency of this node
    def set_dep(self, dep):
        """"""

        self._dep = dep
        self.set_label()

    #=====================================================================
    # Set the dependency labels of all the leaf nodes
    def set_deps(self, deps):
        """"""

        preterminals = self.preterminals()
        for preterminal in preterminals:
            if re.match('\w', preterminal._cat[0]):
                preterminal.set_dep(deps.pop(0))

    #=====================================================================
    # Create a list of tuples from the preterminals
    def to_tuples(self):
        """"""

        for preterminal in self.preterminals():
            yield (preterminal[0], preterminal.category(), preterminal.dependency())

    #=====================================================================
    # Get the last preterminal
    def _get_last_preterm(self):
        """"""

        if self._preterm:
            return self
        else:
            return self[-1]._get_last_preterm()

    #=====================================================================
    # Pop the first contractables
    def _pop_first_contractable(self):
        """"""

        if self._preterm:
            if self[0] in _contractables:
                return self
            else:
                return None
        else:
            first_contractable = self[0]._pop_first_contractable()
            if self[0] == first_contractable or len(self[0]) == 0:
                self.children.pop(0)
                self.pop(0)
            return first_contractable

    #=====================================================================
    # Merge contractables
    def contract(self):
        """"""

        for child in self:
            if isinstance(child, DependencyTree):
                child.contract()
        i = len(self) - 2
        while i >= 0:
            child = self[i]
            last_preterm = child._get_last_preterm()
            j = i + 1
            while j < len(self):
                next_child = self[j]
                first_contractable = next_child._pop_first_contractable()
                if first_contractable is not None:
                    # Merge their cats/leaves
                    last_preterm._cat += '+'+first_contractable.category()
                    last_preterm[0] += first_goeswith[0]
                    last_preterm.children[0] += first_goeswith.children[0]
                    # Disown empty children
                    if len(next_child) == 0:
                        self.pop(j)
                    else:
                        break
                else:
                    break
            i -= 1

    #=====================================================================
    # Basically, read the output of the stanford parser
    @classmethod
    def fromstring(cls, s):
        """"""

        cTree, dGraph = s.split('\n\n')
        dTree = Tree.fromstring(cTree)
        dTree = DependencyTree.convert(dTree)
        deps = []
        dGraph = dGraph.split('\n')
        lastWord = ''
        for dep in dGraph:
            try:
                dep, thisWord = re.match('(.+?)\(.*?, (.*?)\)', dep).groups()
                if thisWord != lastWord:
                    deps.append(dep)
                    lastWord = thisWord
            except:
                print('')
        dTree.set_deps(deps)
        return dTree

    #=====================================================================
    # Convert between different subtypes of Dependency Trees
    @classmethod
    def convert(cls, tree):
        """
        Convert a tree between different subtypes of Tree.  ``cls`` determines
        which class will be used to encode the new tree.

        :type tree: Tree
        :param tree: The tree that should be converted.
        :return: The new Tree.
        """

        if isinstance(tree, Tree):
            children = [cls.convert(child) for child in tree]
            if isinstance(tree, DependencyTree):
                return cls(tree._cat, children, tree._dep)
            else:
                return cls(tree._label, children)
        else:
            return tree

    #=====================================================================
    # Copy the tree
    def copy(self, deep=False):
        """"""

        if not deep:
            return type(self)(self._cat, self, dep=self._dep)
        else:
            return type(self).convert(self)



from ..imports import *
from .txtparsing import *


#####
# PARSE LINE UNITS
#####


def divide_parse_units(dfsyll,nsyll=None,linebreaks=LINEBREAKS,phrasebreaks=PHRASEBREAKS,**kwargs):
    # reverse
    if linebreaks and not phrasebreaks:
        dfsyll['unit_i']=divide_by_line_only(dfsyll)
    elif phrasebreaks and not linebreaks:
        dfsyll['unit_i']=divide_by_phrase_only(dfsyll)
    elif phrasebreaks and linebreaks:
        dfsyll['unit_i']=divide_by_phrase_and_line(dfsyll)
    
    dfsyll['unit_i']=list(divide_overflow(dfsyll))
    dfsyll['unit_i']=dfsyll['unit_i'].rank(method='dense').apply(int)


def divide_by_line_only(dfwords): return divide_by(dfwords,by=['line_i'])
def divide_by_phrase_only(dfwords): return divide_by(dfwords,by=['sent_i','sentpart_i'])
def divide_by_phrase_and_line(dfwords): return divide_by(dfwords,by=['sent_i','sentpart_i','line_i'])
def divide_by(dfwords, by=['sent_i','sentpart_i','line_i']):
    o=[]
    keynow=None
    unit_i=0
    inp = (dfwords[k] for k in by)
    for rowkey in zip(*inp):
        if rowkey!=keynow: unit_i+=1
        o.append(unit_i)
        keynow=rowkey
    return o

        

def get_nsyll(df):
    return len(df[df.word_ipa_i==1])

def divide_overflow(dfsyll,max_nsyll=MAX_SYLL_IN_PARSE_UNIT):
    for ui,unitdf in dfsyll.groupby('unit_i'):
        nsyll=get_nsyll(unitdf)
        if nsyll<=max_nsyll:
            for n in range(len(unitdf)):
                yield ui
        else:
            newui=[]
            nsyllnow=0
            ui2=2
            gl=list(unitdf.groupby(['sent_i','word_i']))
            for wi,wdf in reversed(gl):
                nsyll_word=get_nsyll(wdf)
                for si in range(len(wdf)): newui+=[ui + (1/ui2)]
                nsyllnow+=nsyll_word
                if nsyllnow>=max_nsyll:
                    ui2+=1
                    nsyllnow=0
            yield from reversed(newui)


###
# Iterating over positions
###


def getlenparse(l): return sum(len(x) for x in l)

def iter_mpos(nsyll, starter=[], pos_types=None, maxS=METER_MAX_S, maxW=METER_MAX_W):
    if pos_types is None:
        wtypes = ['w'*n for n in range(1,maxW+1)]
        stypes = ['s'*n for n in range(1,maxS+1)]
        pos_types=[[x] for x in wtypes + stypes]
        
    news=[]
    for pos_type in pos_types:
        if starter and starter[-1][-1]==pos_type[0][0]: continue
        new = starter + pos_type
        # if starter: print(starter[-1][-1], pos_type[0][0], new)
        #if not is_ok_parse(new): continue
        if getlenparse(new)<=nsyll:
            news.append(new)
    
    # news = battle_parses(news)
    if news: yield from news
    # print('\n'.join('|'.join(x) for x in news))
    for new in news: yield from iter_mpos(nsyll, starter=new, pos_types=pos_types)

POSSPARSED={}
def get_poss_parses(n,maxS=METER_MAX_S,maxW=METER_MAX_W):
    global POSSPARSED

    key=(n,maxS,maxW)
    if not key in POSSPARSED:    
        dfpars = pd.DataFrame([
            dict(
                npos=len(x),
                nsyll=getlenparse(x),
                meter='|'.join(x) + '|',
            )
            for x in iter_mpos(n)
            if getlenparse(x)<=n
        ])
        POSSPARSED[key]=dfpars
    return POSSPARSED[key]

# def parse_combo(dfcombo_orig,min_nsyll=4,**kwargs):
#     dfcombo=dfcombo_orig[dfcombo_orig.word_ipa!=""]
#     bounded={}
#     dfpars=get_poss_parses(len(dfcombo))
#     scored={}
#     for nsyll,nsylldf in sorted(dfpars.groupby('nsyll')):
#         if nsyll<min_nsyll: continue
#         scored={}
#         for meter in nsylldf.meter:
#             exclude=False
#             for bmeter in bounded:
#                 if meter.startswith(bmeter):
#                     exclude=True
#                     break
#             if exclude: continue
#             scored[meter] = apply_constraints(dfcombo, meter, **kwargs)
#         for mtr1 in scored:
#             for mtr2 in scored:
#                 if mtr1>=mtr2: continue
#                 s1=scored[mtr1].sum()
#                 s2=scored[mtr2].sum()
#                 s1_ever_better_than_s2 = any(s1<s2)
#                 s2_ever_better_than_s1 = any(s2<s1)
#                 if s1_ever_better_than_s2 and not s2_ever_better_than_s1:
#                     # s1 bounds s2
#                     bounded[mtr2]=mtr1
#                 elif s2_ever_better_than_s1 and not s1_ever_better_than_s2:
#                     # s2 bounds s1
#                     bounded[mtr1]=mtr2
    
#     o=[]
#     for mtr,mdf in scored.items():
#         mdf['slot_meter']=list(mtr.replace('|',''))
#         odf=mdf.join(dfcombo_orig[[]],how='outer')
#         odf['parse']=format_parse_str(mtr)
#         o.append(odf)
#     return concatt(o)

def parse_combo(dfcombo_orig,min_nsyll=None,**kwargs):
    dfcombo=dfcombo_orig[dfcombo_orig.word_ipa!=""]
    bounded={}
    dfpars=get_poss_parses(len(dfcombo))
    scored={}
    for nsyll,nsylldf in sorted(dfpars.groupby('nsyll')):
        if min_nsyll and nsyll<min_nsyll: continue
        scored={}
        for meter in nsylldf.meter:
            exclude=False
            for bmeter in bounded:
                if meter.startswith(bmeter):
                    exclude=True
                    break
            if exclude: continue
            scored[meter] = apply_constraints(dfcombo, meter, **kwargs)
        for mtr1 in scored:
            for mtr2 in scored:
                if mtr1>=mtr2: continue
                if mtr1[-2:]!=mtr2[-2:]: continue
                s1=scored[mtr1].sum()
                s2=scored[mtr2].sum()
                s1_ever_better_than_s2 = any(s1<s2)
                s2_ever_better_than_s1 = any(s2<s1)
                if s1_ever_better_than_s2 and not s2_ever_better_than_s1:
                    # s1 bounds s2
                    # eprint(mtr1,'bounds',mtr2)
                    bounded[mtr2]=mtr1
                elif s2_ever_better_than_s1 and not s1_ever_better_than_s2:
                    # s2 bounds s1
                    # eprint(mtr2,'bounds',mtr1)
                    bounded[mtr1]=mtr2
    
    o=[]
    for mtr,mdf in scored.items():
        mdf['slot_meter']=list(mtr.replace('|',''))
        odf=mdf.join(dfcombo_orig[[]],how='outer')
        odf['parse']=format_parse_str(mtr)
        o.append(odf)
    return concatt(o)

def format_parse_str(pstr):
    return pstr.replace('|','').replace('s','S')
    # if not pstr: return ''
    # # if not pstr.startswith('|'): pstr='|'+pstr
    # # if not pstr.endswith('|'): pstr=pstr+'|'
    # if pstr.endswith('|'): pstr=pstr[:-1]
    
    # return ' '.join(
    #     f'{x:^2}' for x in pstr.replace('s','S').split('|')
    # )

def get_constraints(constraints=DEFAULT_CONSTRAINTS):
    return [
        CONSTRAINTD.get(cname) if type(cname)==str else cname
        for cname in constraints
        if type(cname)!=str or cname in CONSTRAINTD
    ]


def apply_constraints(df, mtr, constraints=DEFAULT_CONSTRAINTS, allow_partial=False, **kwargs):
    if type(mtr)==str:
        mtr=mtr.replace('|','').lower()
        is_s=np.array([float(x=='s') for x in mtr])
    else:
        is_s=mtr

    df=df.iloc[:len(is_s)]
    odf=pd.DataFrame(index=df.index)
    
    for cnstr in get_constraints(constraints):
        o=cnstr(df, is_s)
        if not allow_partial: o=[1.0 if x>0 else 0.0 for x in o]
        odf['*' + cnstr.__name__] = o

    return odf



###
# PARSING UNITS
###

def format_syll_str_parse(syll, mtr, total):
    # if total:
        # syll=format(FACE_ITAL, syll)
    # else:
        # syll=format(STYLE_SANS, syll)
    if mtr=='w' and not total:
        return format(FACE_PLAIN | STYLE_SANS, syll)
    elif mtr=='w' and total:
        return format(FACE_ITAL | STYLE_SANS, syll)
    elif mtr=='s' and not total:
        return format(FACE_BOLD | STYLE_SANS, syll)
    elif mtr=='s' and total:
        return format(FACE_BOLD|FACE_ITAL | STYLE_SANS, syll)
    return syll

def format_word_parse_str(dfparse):
    o=[]
    for wi,wdf in dfparse.groupby('word_i'):
        word_str=wdf.word_str.iloc[0]
        word_ispunc=not any(x.isalpha() for x in word_str)
        word_fmt = ''.join(wdf.syll_str_parse)
        if not word_ispunc: o.append(' ')
        o.append(word_fmt)
    return ''.join(o).strip()

def parse_unit_combos(
        dfunit,
        index=False,
        **kwargs):
    o=[]
    for ci,dfcombo in enumerate(iter_combos(dfunit)):
        # print(ci,'...')
        try:
            ok_parses = parse_combo(dfcombo,**kwargs)
            dfcombo_join = dfcombo[[col for col in dfcombo.columns if col in set(LINEKEY)]]
            ok_parses2 = ok_parses.join(dfcombo_join)
            o.append(ok_parses2)
        except KeyError as e:
            pass
    odf=pd.concat(o) if len(o) else pd.DataFrame()
    if len(odf):
        odf=final_bounding(odf)
        unit_i=dfunit.unit_i.iloc[0]
        odf['unit_i']=unit_i
        parse_il=list(set(list(zip(odf.parse, odf.combo_i))))
        parse_id=dict((x,i+1) for i,x in enumerate(parse_il))
        odf['parse_i']=[parse_id[tup] for tup in zip(odf.parse, odf.combo_i)]
        odf=rank_parses(odf)
        #odf['syll_str_parse']=[x.upper() if y=='s' else x.lower() for x,y in zip(odf.syll_str, odf.slot_meter)]
        # odf['syll_str_parse']=['*'+x if y else x for x,y in zip(odf.syll_str_parse, odf['*total'])]
        # odf['syll_str_parse']=['*'+x if y else x for x,y in zip(odf.syll_str, odf['*total'])]
        # odf['syll_str_parse']=[format_syll_str_parse(xsyll,xmtr) for xsyll,xmtr in zip(odf.syll_str_parse, odf.slot_meter)]
        odf['syll_str_parse']=[format_syll_str_parse(xsyll,xmtr,xtotal) for xsyll,xmtr,xtotal in zip(odf.syll_str, odf.slot_meter, odf['*total'])]

        parse_str_d={}
        for i,pdf in odf.groupby('parse_rank'):
            parse_str=format_word_parse_str(pdf)
            parse_str_d[i]=parse_str
        odf['parse_str']=odf.parse_rank.apply(lambda x: parse_str_d.get(x,''))
        odf=setindex(odf) if index else resetindex(odf)
    return odf

def final_bounding(parses):
    groups = [g for i,g in parses.groupby(['combo_i','parse'])]
    for g in groups: g.bounded=False
    for i,g in enumerate(groups):
        for ii,gg in enumerate(groups):
            if i>=ii: continue
            bounds = df_bounds_df(g,gg)
            if bounds is True:
                gg.bounded=True
            elif bounds is False:
                g.bounded=True
    return pd.concat(g for g in groups if not g.bounded)

def df_bounds_df(df1,df2):
    cdf1=df1[[c for c in df1.columns if c.startswith('*')]]
    cdf2=df2[[c for c in df2.columns if c.startswith('*')]]
    s1=cdf1.sum()
    s2=cdf2.sum()
    s1_ever_better_than_s2 = any(s1<s2)
    s2_ever_better_than_s1 = any(s2<s1)
    if s1_ever_better_than_s2 and not s2_ever_better_than_s1:
        return True
    elif s2_ever_better_than_s1 and not s1_ever_better_than_s2:
        return False
    return None
    

def rank_parses(odf,**kwargs):
    odf[TOTALCOL]=odf[[col for col in odf.columns if col.startswith('*') and col!=TOTALCOL]].sum(axis=1)
    odf['is_troch']=odf['parse'].apply(lambda x: int(x and x[0].lower()=='s'))
    newrankdf=odf[['parse_i',TOTALCOL,'is_troch']].groupby(['parse_i','is_troch']).sum().reset_index().sort_values([TOTALCOL,'is_troch'])
    newrankdf['ranks_int']=list(range(len(newrankdf)))
    # newrankdf['ranks_int']+=1
    newrank=dict(zip(newrankdf.parse_i, newrankdf.ranks_int))
    # display(newrankdf)
    odf['parse_rank']=odf.parse_i.apply(lambda x: int(newrank.get(x,0))+1)
    odf['num_parses']=odf['parse_rank'].max()
    # odf=odf.drop('parse_i',1)
    # odf=odf.drop('is_troch',1)
    # display(odf)
    return odf


def to_lines(dfparses,totalcol=TOTALCOL,rankcol=PARSERANKCOL, agg=sum, only_best=False, **kwargs):
    # parses
    rparses=resetindex(dfparses)
    if only_best: rparses=rparses[rparses.parse_rank==1]
    rcols=set(rparses.columns)
    pkcols = [c for c in PARSELINEKEY if c in rcols]
    # sum by actual data lines
    valcols = [
        c for c in rparses.select_dtypes('number').columns
        if not c in set(pkcols)
        if not c.endswith('_i')
        if not c.endswith('_ii')
        if not c in {'index','level_0'}
    ]
    aggfunc = dict((vc,sum) for vc in valcols)
    medians = [col for col in rcols if 'num_' in col or col=='num_parses']
    for mx in medians:
        if mx in rcols:
            aggfunc[mx]=np.median
    # lgby=['para_i','unit_i','sent_i','sentpart_i','line_i']
    lgby=['para_i','unit_i']
    lgby+=['parse_rank','parse','parse_str']
    # if not only_best:
    #     lgby+=['parse_rank','parse','parse_str']
    # else:
    #     aggfunc['parse_str']='first'
    #     aggfunc['parse']='first'
    #     aggfunc['parse_str']='first'
        # aggfunc['parse']='first'
    lgby=[col for col in lgby if col in set(rparses.columns)]
    try:
        linesums = rparses.groupby(lgby).agg(aggfunc)
    except Exception as e:
        print('!!',e)
        return pd.DataFrame()
    
    linesums=linesums[sorted(linesums.columns)]
    if TOTALCOL in set(linesums.columns):
        linesums=linesums[[TOTALCOL] + [col for col in linesums.columns if col!=TOTALCOL]]
    return linesums
    


accentd={
    'a':'á',
    'e':'é',
    'i':'í',
    'o':'ó',
    'u':'ú',
    'y':'ý',
}
for k,v in list(accentd.items()): accentd[k.upper()]=v.upper()
        
def parse_markdown(dfparse,accent_stresses=False):
    dfp=resetindex(dfparse)
    dfp=dfp[dfp.parse_rank==1]
    mdline=[]
    for l,dflp in sorted(dfp.groupby('unit_i')):
        mdrow=[]
        for w,dfword in sorted(dflp.groupby('word_i')):
            mdword=[]
            for s,dfsyll in sorted(dfword.groupby('syll_i')):
                row=dfsyll.iloc[0]
                #wpref,w,wsuf = split_punct(row.syll_str)
                md=w=row.syll_str
                
                if row.syll_stress in {"P","S"}:
                    if accent_stresses:
                        w2=[]
                        accented=False
                        numvowels=len([y for y in w if y in accentd])
                        for x in w:
                            if not accented and x in accentd:
                                if numvowels<2 or x!='y':
                                    x=accentd[x]
                                    accented=True
                            w2+=[x]
                        md=w=''.join(w2)
                    else:
                        md=f'<b>{md}</b>'
                    
                if row.slot_meter=='s': md=f'<u>{md}</u>'
                if row['*total']>0:
                    md=f'<font style="color:darkred">{md}</font>'
                #md=f'{wpref}{md}{wsuf}'
                is_punc=not any(x.isalpha() for x in row.syll_str)
                mdword+=[md]
            # mdrow.append('.'.join(mdword))
            mdrow.append(''.join(mdword))
            if not is_punc: mdrow[-1]=' '+mdrow[-1]
        mdline.append(''.join(mdrow))
    
    # spltr='\n * '
    # o=spltr + spltr.join(mdline)#.replace('</u> <u>',' ')
    o=' | '.join(mdline)#.replace('</u> <u>',' ')
    return o



# accentd={
#     'a':'á',
#     'e':'é',
#     'i':'í',
#     'o':'ó',
#     'u':'ú',
#     'y':'ý',
# }
# for k,v in list(accentd.items()): accentd[k.upper()]=v.upper()

# def show_parses(parses_bysyll):
#     try:
#         from IPython.display import HTML
#         return HTML(show_parses_md(parses_bysyll))
#     except ImportError:
#         return show_parses_txt(parses_bysyll)
    
# def show_parse_md(dfparse):
# #     print(dfparse.reset_index().line_i.nunique())
# #     print(dfparse.reset_index().linepart_i.nunique())
#     dfp=dfparse.reset_index()
#     mdline=[]
#     for l,dflp in sorted(dfp.groupby('linepart_i')):
#         mdrow=[]
#         #display(dflp)
        
#         for w,dfword in sorted(dflp.groupby('word_i')):
#             mdword=[]
#             for s,dfsyll in sorted(dfword.groupby('syll_i')):
#                 row=dfsyll.iloc[0]
#                 wpref,w,wsuf = split_punct(row.syll_str)
                
#                 if row.is_stressed:
#                     w2=[]
#                     accented=False
#                     numvowels=len([y for y in w if y in accentd])
#                     for x in w:
#                         if not accented and x in accentd:
#                             if numvowels<2 or x!='y':
#                                 x=accentd[x]
#                                 accented=True
#                         w2+=[x]
#                     w=''.join(w2)
                    
#                 md=w#f'{row.syll_str}'
#                 if row.is_s: md=f'<b>{md}</b>'
#                 #if row.is_s: md=f'<u>{md}</u>'
#                 #if row.is_w: md=f'<i>{md}</i>'
#                 #if row.is_stressed: md=f'<i>{md}</i>'
#                 if row['*total']>0:
#                     # md=f'<span style="color:darkred">{md}</span>'
#                     md=f'<font style="color:darkred">{md}</font>'
#                     # md=f'<i>{md}</i>'
#                 md=f'{wpref}{md}{wsuf}'
#                 mdword+=[md]
#             mdrow.append(''.join(mdword))
#         mdline.append(''.join(mdrow))
    
#     # spltr='\n * '
#     # o=spltr + spltr.join(mdline)#.replace('</u> <u>',' ')
#     o=' | '.join(mdline)#.replace('</u> <u>',' ')
#     return o

# def get_best_parse(dfparse):
#     for i,g in dfparse.groupby(['parse_rank','is_troch','parse_i']): break
#     return g

# def get_best_parses(dfparses):
#     o=[get_best_parse(g) for i,g in dfparses.groupby(['stanza_i','line_i','linepart_i'])]
#     return pd.concat(o) if len(o) else pd.DataFrame()

# def show_parse(dfparse,only_best=True):
#     dfparse=get_best_parses(dfparse)

#     try:
#         from IPython.display import HTML
#         return HTML(show_parse_md(dfparse))
#     except ImportError:
#         return show_parse_txt(dfparse)
    

# def show_parses_md(parses_bysyll):
    
#     grps=sorted(parses_bysyll.reset_index().groupby(
#         ['stanza_i','line_i']#,'linepart_i','parse_rank']    
#     ))
#     mdl=[show_parse_md(dfparse) for i,dfparse in grps]
#     omd='\n<br/>'.join(mdl)
#     return omd
    
 
# def info_parses(dfparses_or_path,lcols=['stanza_i','line_i','linepart_i'],min_num_syll=4):
#     infod={}

#     try:
#         if type(dfparses_or_path)==str:
#             if os.path.exists(dfparses_or_path):
#                 dfparses=read_df(dfparses_or_path)
#             else:
#                 return {}
#         else:
#             dfparses=dfparses_or_path    
        
#         # byline
#         allcols=set(dfparses.columns)|set(dfparses.index.names)
#         if 'syll_i' in allcols: dfparses=to_lines(dfparses)
#         allcols=set(dfparses.columns)|set(dfparses.index.names)
#         if 'parse_num_syll' not in allcols: return infod
#         dfparses=dfparses.query(f'parse_num_syll>={min_num_syll}')


#         infod['num_lines_total']=dfparses.groupby(['stanza_i','line_i','linepart_i']).size().shape[0]
        
#         # avg num sylls
#         num_sylls = dfparses.groupby(lcols).parse_num_syll.mean()
#         num_monosylls = dfparses.groupby(lcols).linepart_num_monosyll.mean()
#         infod['num_sylls_avg'] = num_sylls.mean()

#         infod['num_monosylls_avg'] = num_monosylls.mean()
#         num_monosylls_per_10syll = num_monosylls / num_sylls * 10
#         infod['num_monosylls_per_10syll_avg'] = num_monosylls_per_10syll.mean()
        

#         # avg num parses per line?
#         num_parses = dfparses.groupby(lcols).size()# / dfparses.groupby(lcols).parse_num_syll.mean()
#         infod['num_parses_avg'] = num_parses.mean()
#         num_parses_per_10syll = num_parses / num_sylls * 10
#         infod['num_parses_per_10syll_avg'] = num_parses_per_10syll.mean()
        
        
#         # avg viols?
#         vkeys=[col for col in dfparses.columns if col.startswith('*')]
#         for vk in vkeys:
#             vk2='' if vk=='*total' else f'__{vk}'
#             num_viols = dfparses.groupby(lcols)[vk].mean()
#             infod[f'num_viols_avg{vk2}'] = num_viols.mean()
#             num_viols_per_10syll = num_viols /num_sylls * 10
#             infod[f'num_viols_per_10syll_avg{vk2}'] = num_viols_per_10syll.mean()

#         return infod
#     except KeyError:
#         return infod
    
# def show_info_parses(infod):
#     return f'''
# * Metrical friction: **{infod["num_viols_per_10syll_avg"]:.02f}** viols per 10 sylls (**{infod["num_viols_avg"]:.02f}** per parse)
# * Metrical ambiguity: **{infod["num_parses_per_10syll_avg"]:.02f}** parses per 10 sylls (**{infod["num_parses_avg"]:.02f}** per parse)
# '''




# sys imports
import os,sys
import warnings
warnings.filterwarnings('ignore')
from string import punctuation

def is_interactive():
    import __main__ as main
    return not hasattr(main, '__file__')

IS_INTERACTIVE=is_interactive()

if IS_INTERACTIVE:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm

import pandas as pd,numpy as np,random,json,pickle,shutil
from collections import defaultdict,Counter
import subprocess,multiprocessing as mp
# mp.set_start_method('spawn')
# mp.set_start_method('fork')
from pprint import pprint
from itertools import product
pd.options.display.max_columns=False
import re,nltk
import logging
from functools import partial
import diskcache as dc
import requests
from ftfy import fix_text
from IPython.display import Markdown


# constants
ENGINE_PROSODIC='prosodic'
ENGINE_CADENCE='cadence'
ENGINE=ENGINE_PROSODIC
DEFAULT_PARSE_MAXSEC=30
DEFAULT_LINE_LIM=None
DEFAULT_PROCESSORS={'tokenize':'combined'}
MAX_SYLL_IN_PARSE_UNIT=14
MIN_SYLL_IN_PARSE_UNIT=None

MIN_WORDS_IN_PHRASE=2
MAX_WORDS_IN_PHRASE=15
SEPS_PHRASE=set(',:;–—()[].!?"“”’‘')
SEP_STANZA='\n\n'
SEP_PARA='\n\n'
SEP_LINE='\n'
DEFAULT_LANG='en'
PATH_HERE=os.path.abspath(os.path.dirname(__file__))
PATH_CODE=PATH_HERE
PATH_REPO=os.path.abspath(os.path.join(PATH_CODE,'..'))
PATH_HOME=os.path.join(os.path.expanduser('~'),'cadence_data')
PATH_DATA=os.path.join(PATH_HOME,'data')
PATH_TXTS=os.path.join(PATH_HOME,'txts')
if not os.path.exists(PATH_DATA): os.makedirs(PATH_DATA)
if not os.path.exists(PATH_TXTS): os.makedirs(PATH_TXTS)
PATH_DB_SYLLABIFY=os.path.join(PATH_DATA,'syllabified.dc')
# PATH_DATA=os.path.join(PATH_REPO,'data')
DATA_URL='https://www.dropbox.com/s/fywmqrlpemjf43c/data_cadence.zip?dl=1'


PATH_NOTEBOOKS=os.path.join(PATH_REPO,'notebooks')
PATH_IPA_FEATS=os.path.join(PATH_DATA,'data.feats.ipa.csv')
PATH_DB_SCAN=os.path.join(PATH_DATA,'db_scan.dc')
PATH_DB_PARSE=os.path.join(PATH_DATA,'db_parse.dc')
PATH_DB=os.path.join(PATH_DATA,'db.dc')
DBSEP='____'

METER_MAX_S=2
METER_MAX_W=2


LINEBREAKS=False
PHRASEBREAKS=True


INCL_ALT=True
DEFAULT_NUM_PROC=1
#mp.cpu_count()//2# - 1

DEFAULT_METER='default_english'
COLS_JOINER=['para_i','sent_i','word_i']

KEEP_BEST=1
SBY=csby=['combo_i','word_i','syll_i']
PARSERANKCOL='parse_rank'

LINEKEY=[
    'id_author',
    'id',
    'para_i','stanza_i',
    'para_start_char','para_end_char',


    'unit_i',
    'parse_rank','is_troch','parse_i','parse','parse_str',

    'sent_i'] + [
        f'sent_depth{i}' for i in range(1,10)
    ] + [
        # 'word_i',
    'sentpart_i',


    'linepart_i','linepart_str','linepart_end_str',

    'line_i','line_str',

    
    'combo_stress','combo_ipa','combo_i',
    'slot_i','slot_meter','syll_str_parse',
    
    'parse_is_bounded','parse_bounded_by',
    'parse_pos_i','parse_pos',
    
    'word_i',
    'word_pref','word_str','word_tok','word_ipa_i','word_ipa',
    'word_lemma','word_upos','word_xpos','word_deprel','word_head',
    'word_constituency',
    'word_start_char','word_end_char',
    
    'syll_i','combo_syll_i','syll_str','syll_ipa','syll_stress','syll_weight',
    
    'parse_syll_i','parse_syll',


]
PARSELINEKEY = LINEKEY[:LINEKEY.index('parse_pos_i')]
PARSESYLLKEY=LINEKEY
TOTALCOL='*total'

DEFAULT_CONSTRAINTS = [
    'w_peak',
    'w_stressed',
    's_unstressed',
    'unres_across',
    'unres_within',
]
DEFAULT_CONSTRAINTS_P = [
    'w_peak',
    'w_stressed',
    's_unstressed',
    'unres_across',
    'unres_within',
    'w_stressed_t'
]

NUM_BOUNDED_TO_STORE = 10

constraint_names_in_prosodic = {
    '*f-res':'footmin-f-resolution',
    '*s/unstressed':'stress.s=>-u',
    '*w-res':'footmin-w-resolution',
    '*w/peak':'strength.w=>-p',
    '*w/stressed':'stress.w=>-p',
    '*s/trough':'strength.s=>-u',
}
constraint_names_from_prosodic = dict((v,k) for k,v in constraint_names_in_prosodic.items())


txt="""In Xanadu did Kubla Khan
A stately pleasure-dome decree:
Where Alph, the sacred river, ran
Through caverns measureless to man
   Down to a sunless sea.
"""
line=txt.split('\n')[0]

APPLY_POSTHOC_CONSTRAINTS=False
WORD_TOKDF={}


URLS=dict(
    prideprej='https://www.gutenberg.org/files/1342/1342-0.txt'
)

NLP_PARSE_POSTAG=True
NLP_PARSE_TOKENIZE=True
NLP_PARSE_CONSTITUENCY=True
NLP_PARSE_DEPPARSE=True
SHUFFLE_PARAS=False
LIM_PARAS=None

joyce_path = os.path.join(PATH_TXTS,'joyce_oxen.txt')
joyce_s = s = 'Stately, plump Buck Mulligan came from the stairhead, bearing a bowl of lather on which a mirror and a razor lay crossed.'
joyce_para="""
Valuing himself not a little upon his elegance, being indeed a proper man of his person, this talkative now applied himself to his dress with animadversions of some heat upon the sudden whimsy of the atmospherics while the company lavished their encomiums upon the project he had advanced. The young gentleman, his friend, overjoyed as he was at a passage that had befallen him, could not forbear to tell it his nearest neighbour. Mr Mulligan, now perceiving the table, asked for whom were those loaves and fishes and, seeing the stranger, he made him a civil 290bow and said, Pray, sir, was you in need of any professional assistance we could give? Who, upon his offer, thanked him very heartily, though preserving his proper distance, and replied that he was come there about a lady, now an inmate of Horne's house, that was in an interesting condition, poor lady, from woman's woe (and here he fetched a deep sigh) to know if her happiness had yet taken place. Mr Dixon, to turn the table, took on to ask of Mr Mulligan himself whether his incipient ventripotence, upon which he rallied him, betokened an ovoblastic gestation in the prostatic utricle or male womb or was due, as with the noted physician, 403Mr Austin Meldon, to a wolf in the stomach. For answer Mr Mulligan, in a gale of laughter at his 384smalls, smote himself bravely below the diaphragm, exclaiming with an admirable droll mimic of Mother Grogan (the most excellent creature of her sex though 'tis pity she's a trollop): There's a belly that never bore a bastard. This was so happy a conceit that it renewed the storms of mirth and threw the whole room into the most violent agitations of delight. The spry rattle had run on in the same vein of mimicry but for some larum in the antechamber.
"""


en_prose_taylor="""Prayer is the peace of our spirit, the stillness of our thoughts, the evenness of recollection, the seat of meditation, the rest of our cares and the calm of our tempest. Prayer is the issue of a quiet mind, of untroubled thoughts; it is the daughter of charity and the sister of meekness; and he that prays to God with an angry — that is a troubled and discomposed —spirit, is like him that retires into a battle to meditate and sets up his closet in the outquarters of an army and chooses a frontier garrison to be wise in. Anger is a perfect alienation of the mind from prayer, and therefore is contrary to that attention which presents our prayers in a right line to God. For so have I seen a lark rising from his bed of grass, soaring upwards and singing as he rises and hopes to get to Heaven and climb above the clouds; but the poor bird was beaten back with the loud sighings of an eastern wind and his motion made irregular and inconstant, descending more at every breath of the tempest than it could recover by the vibration and frequent weighing of his wings; till the little creature was forced to sit down and pant and stay till the storm was over; and then it made a prosperous flight and did rise and sing as if it had learned music and motion from an angel as he passed sometimes through the air about his ministries here below. So is the prayer of a good man; when his affairs have required business, and his business was matter of discipline, and his discipline was to pass upon a sinning person, or had a design of charity, his duty met with infirmities of a man and anger was its instrument, and the instrument became stronger than the prime agent and raised a tempest and overruled the man; and then his prayer was broken and his thoughts troubled."""
en_prose_taylor2="""For so have I seen a lark rising from his bed of grass, soaring upwards and singing as he rises and hopes to get to Heaven and climb above the clouds; but the poor bird was beaten back with the loud sighings of an eastern wind and his motion made irregular and inconstant, descending more at every breath of the tempest than it could recover by the vibration and frequent weighing of his wings; till the little creature was forced to sit down and pant and stay till the storm was over; and then it made a prosperous flight and did rise and sing as if it had learned music and motion from an angel as he passed sometimes through the air about his ministries here below. So is the prayer of a good man; when his affairs have required business, and his business was matter of discipline, and his discipline was to pass upon a sinning person, or had a design of charity, his duty met with infirmities of a man and anger was its instrument, and the instrument became stronger than the prime agent and raised a tempest and overruled the man; and then his prayer was broken and his thoughts troubled."""
en_prose_burke1="""To complain of the age we live in, to murmur at the present possessors of power, to lament the past, to conceive extravagant hopes of the future, are the common dispositions of the greatest part of mankind; indeed the necessary effects of the ignorance and levity of the vulgar. Such complaints and humours have existed in all times; yet as all times have not been alike, true political sagacity manifests itself, in distinguishing that complaint which only characterizes the general infirmity of human nature, from those which are symptoms of the particular distemperature of our own air and season."""




# local imports
from .tools import *
from .langs import *
from .constraints import *
from .parsers import *
from .cadence import *


# check
check_basic_config()







from ..imports import *
# from .english import *
from .english import scan as en_scan, get as en_get
from .english import get_df
SYLLABIFY_DF_D={}

CODE2LANG = {
    'en':en_scan,
}
CODE2LANG_SYLLABIFY={
    'en':en_get,
}

def ipa_to_stress(syl_ipa,numeric=True):
    if not syl_ipa or type(syl_ipa)!=str:
        return np.nan if numeric else ''
    elif syl_ipa[0]=="'":
        return 1.0 if numeric else 'P'
    elif syl_ipa[0]== "`":
        return 0.5 if numeric else 'S'
    else:
        return 0.0 if numeric else 'U'


def get_vowels(sylipa):
    return [
        ipa
        for ipa in sylipa
        if get_d_ipa().get(ipa,{}).get('syll') == True
    ]

def syll_is_dipthong(sylipa):
    return len(get_vowels(sylipa))>1

def syll_ends_with_consonant(sylipa):
    return get_d_ipa().get(sylipa[-1],{}).get('syll')==False

def syll_is_heavy(sylipa):
    return syll_ends_with_consonant(sylipa) or syll_is_dipthong(sylipa)

def ipa_to_weight(sylipa,numeric=True):
    is_heavy = syll_is_heavy(sylipa)
    if is_heavy:
        return 1.0 if numeric else 'H'
    else:
        return 0.0 if numeric else 'L'


def apply_proms_sylls(word_df):
    df=word_df
    cols=set(df.columns)
    if 'syll_ipa' in cols:
        df['prom_stress']=df['syll_ipa'].apply(ipa_to_stress)
        df['prom_weight']=df['syll_ipa'].apply(getweight)
        df['prom_strength']=[
            x
            for i,df_word in df.groupby('word_ipa_i')
            for x in getstrength(df_word.prom_stress)
        ]
        df['is_stressed']=(df['prom_stress']>0).apply(np.int32)
        df['is_unstressed']=(df['prom_stress']==0).apply(np.int32)
        df['is_heavy']=(df['prom_weight']==1).apply(np.int32)
        df['is_light']=(df['prom_weight']==0).apply(np.int32)
        df['is_peak']=(df['prom_strength']==1).apply(np.int32)
        df['is_trough']=(df['prom_strength']==0).apply(np.int32)

SYLL_LD_CACHE={}
def get_syllable_ld(word_str,lang=DEFAULT_LANG,force=False,**kwargs):
    global SYLL_LD_CACHE
    key=(lang,word_str.strip())
    if not key in SYLL_LD_CACHE:
        # print('looking up:',key)
        ld=CODE2LANG_SYLLABIFY[lang](word_str.strip(), **kwargs)
        # is_punc=not any(x.isalpha() for x in word_str)
        if not ld:
            ld=[{
                'word_ipa_i':0,
                'syll_i':0,
                'word_str':word_str,
                'word_tok':to_token(word_str),
                'word_ipa':"",
                'word_nsyll':0,
                'syll_ipa':"",
                'syll_str':word_str,# if is_punc else 
                'word_isfunc':np.nan
            }]
            # pass
        else:
            # tune ups
            for dx in ld:
                syll_ipa=dx['syll_ipa']
                dx['syll_stress']=ipa_to_stress(syll_ipa,numeric=False)
                dx['syll_weight']=ipa_to_weight(syll_ipa,numeric=False)
                dx['prom_stress']=ipa_to_stress(syll_ipa,numeric=True)
                dx['prom_weight']=1.0 if dx['syll_weight']=='H' else 0.0
            
            stresses = dict((dxi,dx.get('prom_stress')) for dxi,dx in enumerate(ld))
            if None not in set(stresses.values()):
                strengths = getstrength(stresses)
                for i,strength in enumerate(strengths):
                    ld[i]['prom_strength']=strength
        SYLL_LD_CACHE[key]=ld
    return [
        {**dx, **{'word_str':word_str}}
        for dx in SYLL_LD_CACHE[key]
    ]



SYLL_DF_CACHE={}
def get_syllable_df(word_str,lang=DEFAULT_LANG,force=False,**kwargs):
    global SYLL_DF_CACHE
    key=(lang,word_str)
    if not key in SYLL_DF_CACHE:
        # print('df looking up')
        odf=pd.DataFrame(get_syllable_ld(word_str,force=force,**kwargs))
        SYLL_DF_CACHE[key]=odf
    return SYLL_DF_CACHE[key]


def syllabify_df(df,**kwargs):
    df=resetindex(df)
    cols=set(df.columns)
    dfsyll=pd.concat(
        get_syllable_df(word_str,index=False,**kwargs)
        for word_str in df.word_str.unique()
    )
    if not len(dfsyll): return pd.DataFrame()
    odf=df.merge(dfsyll,on='word_str',how='left')
    ## mtree
    for col in ['prom_pstress', 'prom_lstress', 'prom_tstress', 'prom_pstrength']:
        if col in cols:
            odf.loc[
                ((odf['prom_stress']==0.0) & (odf['word_nsyll']>1)),
                col
            ]=np.nan
    return odf


def to_lang(lang_code=None):
    global LANGS
    return LANGS[lang_code]

D_IPA=None
def get_d_ipa(fn=PATH_IPA_FEATS):
    global D_IPA
    if D_IPA is None: D_IPA=pd.read_csv(fn).set_index('ipa').T.to_dict()
    return D_IPA



def assign_proms(df):
    if not len(df): return
    # set proms
    df['prom_stress']=pd.to_numeric(df['syll_ipa'].apply(getstress),errors='coerce')
    df['prom_weight']=pd.to_numeric(df['syll_ipa'].apply(getweight),errors='coerce')
    # df['syll_stress']=df.prom_stress.apply(lambda x: {1.0:'P', 0.5:'S', 0.0:'U'}.get(x,''))
    df['syll_weight']=df.prom_weight.apply(lambda x: {1.0:'H', 0.0:'L'}.get(x,''))

    df['prom_strength']=[
        x
        for i,df_word in df.groupby('word_ipa_i')
        for x in getstrength(df_word)
    ]
    df['is_stressed']=(df['prom_stress']>0).apply(np.int32)
    df['is_unstressed']=(df['prom_stress']==0).apply(np.int32)
    df['is_heavy']=(df['prom_weight']==1).apply(np.int32)
    df['is_light']=(df['prom_weight']==0).apply(np.int32)
    df['is_peak']=(df['prom_strength']==1).apply(np.int32)
    df['is_trough']=(df['prom_strength']==0).apply(np.int32)


def to_phons(syll_ipa):
    pass

def to_int(x):
    try:
        xint=int(x)
        if x==xint:
            return xint
    except ValueError:
        pass
    return x

def nice_int(odf):
    for col in odf.columns:
        if col.endswith('_i') or col.startswith('is_'):
            odf[col]=pd.to_numeric(odf[col], errors='coerce', downcast='integer')
    return odf



def line2df(line_txt,
        lang=DEFAULT_LANG,
        sby=['word_i','word_ipa_i','syll_i'],
        incl_alt=INCL_ALT,**y):
    
    func = CODE2LANG.get(lang, CODE2LANG[DEFAULT_LANG] )
    odf=func(line_txt,incl_alt=incl_alt,**y)
    # try:
    #     # odf=odf.sort_values(sby)[
    #     #     sby + [col for col in odf.columns if col not in set(sby)]
    #     # ]
    #     # odf['is_syll']=odf.syll_ipa.apply(lambda x: int(x==''))#[int(x) for x in odf['syll_ipa']!='']
    #     #odf['line_num_syll']=len(odf[(odf.is_syll==1) & (odf.word_ipa_i==1)])
    #     #uniqdf=odf[odf.is_syll==1].drop_duplicates(['word_i','syll_i'])
    #     #odf['line_num_syll']=len(uniqdf)
    # except KeyError as e:
    #     pass
    return odf




# ## langs


# stress
def getstress(sylipa):
    if not sylipa.strip(): return ' '
    if sylipa.startswith("'"): return 1.0#""
    elif sylipa.startswith("`"): return 0.5
    return 0.0

def getstress_str(sylipa):
    x=getstress(sylipa)
    if x==1.0: return 'P'
    if x==0.5: return 'S'
    if x==0.0: return 'U'
    return ''

def getweight(sylipa):
    dipa=get_d_ipa()
    phondata = [dipa[ipastr] for ipastr in sylipa if ipastr in dipa]
    if not len(phondata): return np.nan
    ends_with_cons=phondata[-1]['cons']==True
    has_long_vowel=any(x['long']==True for x in phondata)
    is_dipthong=None #@TODO
    weight=1 if ends_with_cons or has_long_vowel or is_dipthong else 0
    return weight

def getweight_str(sylipa):
    if x==1: return 'H'
    if x==0: return 'L'
    return ''

# def getweight(self,df_syll):
#     ends_with_cons=df_syll.sort_values('phon_i')['phon_cons'].iloc[-1]==True
#     has_long_vowel=any(x==True for x in df_syll['phon_long'])
#     # print(has_long_vowel,ends_with_cons,[df_syll.sort_values('phon_i')['phon_cons'].iloc[-1]])
#     is_dipthong=None #@TODO
#     weight='H' if ends_with_cons or has_long_vowel or is_dipthong else 'L'
#     return [weight for i in range(len(df_syll))]

# def getstrength(df_word):
#     stresses=dict(zip(df_word.syll_i,df_word.prom_stress))
def getstrength(stresses):
    strengths=[]
    for si,(syll_i,syl) in enumerate(sorted(stresses.items())):
        prv=stresses.get(syll_i-1)
        nxt=stresses.get(syll_i+1)
        if nxt!=None and prv!=None:
            if syl>nxt or syl>prv:
                strength=1.0
            elif syl<nxt or syl<prv:
                strength=0.0
            else:
                strength=np.nan
        elif prv==None and nxt!=None:
            if syl>nxt:
                strength=1.0
            elif syl<nxt:
                strength=0.0
            else:
                strength=np.nan
        elif prv!=None and nxt==None:
            if syl>prv:
                strength=1.0
            elif syl<prv:
                strength=0.0
            else:
                strength=np.nan
        elif prv==None and nxt==None:
            strength=np.nan
        else:
            raise Exception("How? -getstrength()")
        #strengths.append(strength)
        strengths.append(strength)
    return strengths




from ..imports import *

"""
All constraints take the entire dataframe
"""

def w_stressed(df, is_s):
    return (1-is_s) * np.array([1.0 if x>0.0 else 0.0 for x in df.prom_stress])
def w_stressed_l(df, is_s):
    return (1-is_s) * np.array([1.0 if x>0.0 else 0.0 for x in df.prom_lstress])
def w_stressed_p(df, is_s):
    return (1-is_s) * np.array([1.0 if x>0.0 else 0.0 for x in df.prom_pstress])
def w_stressed_t(df, is_s):
    return (1-is_s) * np.array([1.0 if x>0.0 else 0.0 for x in df.prom_tstress])

def s_unstressed(df, is_s):
    return is_s * np.array([1.0 if x==0.0 else 0.0 for x in df.prom_stress])
def s_unstressed_l(df, is_s):
    return is_s * np.array([1.0 if x==0.0 else 0.0 for x in df.prom_lstress])
def s_unstressed_p(df, is_s):
    return is_s * np.array([1.0 if x==0.0 else 0.0 for x in df.prom_pstress])
def s_unstressed_t(df, is_s):
    return is_s * np.array([1.0 if x==0.0 else 0.0 for x in df.prom_tstress])


def w_peak(df, is_s):
    return (1-is_s) * df.prom_strength.values
def w_peak_p(df, is_s):
    return (1-is_s) * df.prom_pstrength.values
def s_trough(df, is_s):
    return is_s * (1-df.prom_strength.values)
def s_trough_p(df, is_s):
    return is_s * (1-df.prom_pstrength.values)


def is_disyllab_pos(is_s):
    return np.array([
        float(is_s[i] and ((i and is_s[i-1]) or (i+1<len(is_s) and is_s[i+1])))
        for i in range(len(is_s))
    ])

def unres_across(df,is_s):
    o=[np.nan] # need to start with non-viol since this applies only to second position of disyllab
    for i in range(1,len(df)):
        # for every 2 consecutive sylls...
        s1=is_s[i-1]
        s2=is_s[i]
        if s1!=s2:
            # not disyllabic
            o.append(np.nan)
        else:
            row1,row2=df.iloc[i-1],df.iloc[i]
            if (row1.sent_i, row1.word_i) == (row2.sent_i, row2.word_i):
                # disyllabic position is within not across words
                o.append(np.nan)
            else:
                if s1 and s2:
                    # disyllabic strong position immediately violates
                    # o[-1]=1.0  # prev position too?
                    o.append(1.0)
                    
                else:
                    if not row1.word_isfunc or not row2.word_isfunc:
                        # o[-1]=1.0  # prev position too?
                        o.append(1.0)
                    else:
                        o.append(0.0)
    return o

def unres_within(df,is_s):
    o=[np.nan] # need to start with non-viol since this applies only to second position of disyllab
    for i in range(1,len(df)):
        # for every 2 consecutive sylls...
        s1=is_s[i-1]
        s2=is_s[i]
        if s1!=s2:
            # not disyllabic
            o.append(np.nan)
        else:
            row1,row2=df.iloc[i-1],df.iloc[i]
            if (row1.sent_i, row1.word_i) != (row2.sent_i, row2.word_i):
                # disyllabic position is across not wihtin words
                o.append(np.nan)
            else:
                # disyllabic position within word
                # first position mist be light and stressed
                if not (row1.prom_weight==0 and row1.prom_stress>0):
                    # o[-1]=1.0  # prev position too?
                    o.append(1.0)
                else:
                    o.append(0.0)
    return o



def bad_window(s,badwindow=(1,1)):
    wlen=len(badwindow)
    l=[]
    window=[]
    for x in s:
        window.append(x)
        l.append(int(tuple(window)==badwindow))
        if len(window)>=wlen:window.pop(0)
    return l

def clash(df, is_s):
    is_stressed=[int(x>0) for x in df.prom_stress]
    return no_window(is_stressed,badwindow=(1,1))
def lapse(df, is_s):
    is_stressed=[int(x>0) for x in df.prom_stress]
    return bad_window(is_stressed,badwindow=(0,0,0))





CONSTRAINTL = [
    w_stressed,
    w_stressed_l,
    w_stressed_p,
    w_stressed_t,
    
    s_unstressed,
    s_unstressed_l,
    s_unstressed_p,
    s_unstressed_t,

    w_peak,
    w_peak_p,

    s_trough,
    s_trough_p,
    
    unres_across,
    unres_within,

    clash,
    lapse
]

CONSTRAINTD=dict((func.__name__,func) for func in CONSTRAINTL)


from ..imports import *
from .mtree import recurse_tree

NLPD={}
# badcols={'feats','id','text'}
badcols={'feats','start_char','end_char','id','text','misc','lemma'}

def get_nlp(
        lang=DEFAULT_LANG,
        
        pretokenized=True,
        postag=NLP_PARSE_POSTAG,
        tokenize=NLP_PARSE_TOKENIZE,
        constituency=NLP_PARSE_CONSTITUENCY,
        depparse=NLP_PARSE_DEPPARSE,

        processors=[],
        verbose=False,
        **kwargs
        ):
    global NLPD

    if not processors:
        processors=get_processors(
            constituency=constituency,
            depparse=depparse,
            postag=postag
        )
    
    procstr=','.join(processors)
    
    kwargs=dict(
        lang=lang,
        tokenize_pretokenized=pretokenized,
        processors=procstr,
    )

    key=kwargs_key(kwargs)

    if not key in NLPD:
        # eprint('Loading NLP model:',kwargs)
        
        import stanza
        kwargs2={**dict(verbose=False), **kwargs}
        try:
            NLPD[key] = stanza.Pipeline(**kwargs2)
        except stanza.pipeline.core.ResourcesFileNotFoundError:
            lang = kwargs.get('lang')
            eprint(f'[cadence] Downloading stanza NLP model for language = "{lang}"')
            stanza.download(lang)
            try:
                NLPD[key] = stanza.Pipeline(**kwargs2)
            except stanza.pipeline.core.ResourcesFileNotFoundError:
                raise Exception(f'[cadence] Cannot download stanza model for language = "{lang}"')
        
        NLPD[key].procstr=procstr
        # NLPD[key].processors=processors
    return NLPD[key]




def get_processors(
        tokenize=True,
        postag=True,
        constituency=True,
        depparse=True,
        **kwargs):
    #o=dict(DEFAULT_PROCESSORS)
    o={}
    if tokenize:
        o['tokenize']=''
    if postag:
        o['tokenize']=''
        o['pos']=''
        o['lemma']=''
    if constituency:
        o['tokenize']=''
        o['pos']=''
        o['constituency']=''
    if depparse:
        o['tokenize']=''
        o['pos']=''
        o['lemma']=''
        o['depparse']=''
    return sorted(list(o.keys()))




########
# Tokenize NLP
########


### NLP FUNCS
def get_nlp_doc(doc_ll_or_txt,nlp=None,**kwargs):
    if nlp is None: nlp=get_nlp(**kwargs)
    try:
        doc=nlp(doc_ll_or_txt)
        # fix
        fix_nlp_doc_constituency_bug(doc)
        return doc
    except Exception as e:
        # eprint(f'!! NLP Parser error: {e}')
        pass

COLS_RENAMER_NLP=dict(
    deprel='dep_type',
    head='dep_head',
    upos='pos_upos',
    xpos='pos_xpos',
    lemma='word_lemma',
)


def get_nlp_doc_wordfeat_df(doc,**kwargs):
    if doc is None: return pd.DataFrame()
    ld=[]
    sents=doc.sentences
    sentd_orig=None
    for sent_i, sent in enumerate(sents):
        ## Get Word Info
        for word_i,word in enumerate(sent.tokens):
            feats=word.to_dict()[0]
            statd=dict(
                # (v,feats[k])
                # for k,v in COLS_RENAMER_NLP.items()
                # if k in feats
                (COLS_RENAMER_NLP.get(k,k), feats[k])
                for k in feats
                if k not in badcols
            )
            for feat in feats.get('feats','').split('|'):
                if not feat: continue
                fk,fv=feat.split('=',1)
                statd[f'pos_{fk.lower()}']=fv
            dx={
                'sent_i': sent_i+1,
                'word_i': word_i+1,
                'word_str':word.text,
                **statd
            }
            ld.append(dx)
    return pd.DataFrame(ld).fillna('')

def fix_nlp_doc_constituency_bug(doc,**kwargs):
    sentd_orig={}
    sents = doc.sentences
    for sent_i,sent in enumerate(sents):
        sent.i=sent_i+1
        sentd_orig[get_sent_id_tokens(sent)] = sent
    for sent_i,sent in enumerate(sents):
        if hasattr(sent,'constituency'):
            sent_id_constituency=get_sent_id_constituency(sent)
            if sent_id_constituency not in sentd_orig: continue
            sent_orig=sentd_orig[sent_id_constituency]
            if hasattr(sent_orig,'constituency'):
                sent.constituency,sent_orig.constituency = sent_orig.constituency,sent.constituency
    
    doc.sentences = sents


def get_nlp_doc_constituency_df(doc,**kwargs):
    if doc is None: return pd.DataFrame()
    ## Constituency?
    ld=[]
    sents=doc.sentences
    sentd_orig=None

    depthstrd=Counter()
    def getdepthstr(lvlstr):
        lvlstr=lvlstr.split('-',1)[-1]
        depthstrd[lvlstr]+=1
        return f'{lvlstr}-{depthstrd[lvlstr]}'

    pathseend={}

    for sent_i, sent in enumerate(sents):
        if hasattr(sent,'constituency'):
            senttree=recurse_tree(sent.constituency, node_i=0, path=[])
            for word_i,word_constituency_path in enumerate(senttree):
                word_constituency_path_str='('.join(word_constituency_path)
                word_constituency_path_nolvl=[xx.split('-',1)[-1] for xx in word_constituency_path]
                constituency_depth=len(word_constituency_path)
                dx={
                    'sent_i': sent.i,
                    'word_i': word_i+1,
                    'word_depth':constituency_depth,
                    # 'word_constituency':word_constituency_path_str,
                }
                
                # max_wc_len=10
                # for wci in range(2,len(word_constituency_path)+1):
                #     if wci>max_wc_len: break
                #     wcpath=word_constituency_path[:wci]
                #     pathseen=tuple(wcpath)
                #     if not pathseen in pathseend: pathseend[pathseen]=f'{word_i+1:03}'
                #     newlevel=pathseend[pathseen]
                #     wcpathstr=f'{newlevel}_{"(".join(wcpath)}'
                #     dx[f'sent_depth{wci-1}']=wcpathstr
                ld.append(dx)
    return pd.DataFrame(ld).fillna('')


def get_nlp_feats_df(doc,**kwargs):
    df_feats=get_nlp_doc_wordfeat_df(doc,**kwargs)
    # display(df_feats)

    df_const=get_nlp_doc_constituency_df(doc,**kwargs)
    # display(df_const)

    if len(df_feats) and len(df_const):
        return df_feats.merge(df_const,on=['sent_i','word_i'],how='inner')
    elif len(df_feats):
        return df_feats
    elif len(df_const):
        return dc_const
    return pd.DataFrame()



def get_sent_id_tokens(sent,hash=True):
    o=[tok.text for tok in sent.tokens]
    return tuple(o)
def get_sent_id_constituency(sent,hash=True):
    o=[tok.split()[-1] for tok in str(sent.constituency).split(')') if tok]
    o=[x.replace('(-RRB-',')') for x in o]
    return tuple(o)







from ..imports import *
from .txtparsing import *
from .mtree import *
from .stanzanlp import *
from .metrics import *


def Text(*args,**kwargs):
    return TextModel(*args,**kwargs)

def Prose(*args,**kwargs):
    kwargs={**kwargs, **dict(linebreaks=False, phrasebreaks=True)}
    return Text(*args,**kwargs)
def Verse(*args,**kwargs):
    kwargs={**kwargs, **dict(linebreaks=True, phrasebreaks=False)}
    return Text(*args,**kwargs)
def ProseVerse(*args,**kwargs):
    kwargs={**kwargs, **dict(linebreaks=True, phrasebreaks=True)}
    return Text(*args,**kwargs)

def Para(txt, *args, path_db=PATH_DB, force=False, para_i=None, **kwargs):
    txt=txt.strip()
    _id=get_para_key(txt, kwargs)
    #print('Para Key =',_id)
    if not force:
        with dc.Cache(path_db) as db:
            if _id in db:
                # eprint(f'[cadence] Found paragraph cached in db, loading...')
                pdoc=pickle.loads(from_blosc(db[_id]))
                pdoc.i=para_i
                return pdoc

    pdoc=ParaModel(txt,*args,path_db=path_db,**kwargs)
    pdoc.i=para_i
    return pdoc

def Sent(txt, nlp_sent_obj=None, **kwargs):
    return SentModel(txt, nlp_sent_obj=nlp_sent_obj, **kwargs)


def get_para_key(txt, kwargs):
    return (hashstr(txt), kwargs_key(kwargs))

###
# TEXT
###

class TextModel(object):
    def __init__(self,txt_or_fn,shuffle_paras=SHUFFLE_PARAS,lim_paras=LIM_PARAS,**kwargs):
        self.fn,self.txt=to_fn_txt(txt_or_fn)
        self._kwargs=kwargs
        self._lim_paras=lim_paras
        self._shuffle_paras=shuffle_paras
        self._docs={}

        # paras
        self._paras=dict(
            (para_d['para_i'],para_d)
            for para_d in tokenize_paras_ld(self.txt, **self.kwargs(kwargs))
        )
        self.__paras_i=set(self._paras.keys())
        self._paras_i=sorted(list(self.__paras_i))
        self._paras_i=self.paras_i(shuffle_paras=shuffle_paras,lim_paras=lim_paras,progress=False,**self.kwargs(kwargs))
        self._num_paras=len(self._paras)

    def __repr__(self):
        return f'<Text: “{self.para(1).fsent_str}...” ({self.num_paras} paras)>'
    
    def kwargs(self,kwargs):
        return {**self._kwargs, **kwargs}

    def get(self,para_i,sent_i):
        obj=self.para(para_i)
        if sent_i: obj=obj.sent(sent_i)
        return obj
        

    """
    PARA MANAGEMENT
    """

    @property
    def num_paras(self): return len(self._paras)
    
    def paras_i(self,
            para_i=[],
            shuffle_paras=False,
            reset_paras=False,
            lim_paras=None,
            progress=True,
            desc='Iterating over paragraphs',
            **kwargs):
        
        if para_i:
            para_i_l=[para_i] if type(para_i)==int else list(para_i)
        else:
            para_i_l=self._paras_i
        if reset_paras: para_i_l=sorted(list(self.__paras_i))
        if shuffle_paras: random.shuffle(para_i_l)
        if lim_paras: para_i_l=para_i_l[:lim_paras]
        if progress and len(para_i_l)>1: para_i_l=tqdm(para_i_l,desc=desc)
        return para_i_l
    
    def paras_d(self,*args,**kwargs):
        return [self._paras[para_i] for para_i in self.paras_i(*args,**self.kwargs(kwargs))]    
    
    def paras_df(self,*args,**kwargs):
        return pd.DataFrame(self.paras_d(*args,**self.kwargs(kwargs))).set_index('para_i')
    
    def iter_paras(self,*args,**kwargs):
        for para_i in self.paras_i(*args,**self.kwargs(kwargs)):
            yield self.para(para_i,**self.kwargs(kwargs))
    
    def para(self,para_i=1,**kwargs):
        if not para_i in self._docs:
            if not para_i in self._paras: return
            txt=self._paras[para_i]['para_str']
            #print('Making new paragraph',para_i,self.kwargs(kwargs),'...')
            self._docs[para_i]=Para(txt, para_i=para_i, **self.kwargs(kwargs))
        return self._docs[para_i]
    def paras(self,*args,**kwargs):
        return list(self.iter_paras(*args,**self.kwargs(kwargs)))

    @property
    def num_stanzas(self): return self.num_paras
    def stanzas_i(self,**kwargs): return self.paras_i(**kwargs)
    def stanzas_d(self,**kwargs): return self.paras_d(**kwargs)
    def stanzas_df(self,**kwargs): return self.paras_df(**kwargs)
    def iter_stanzas(self,**kwargs): return self.iter_paras(**kwargs)
    def stanza(self,**kwargs): return self.para(**kwargs)
    def stanzas(self,**kwargs): return self.paras(**kwargs)
    
    #############################################
    ### Data
    #############################################

    def iter_data(self,data_type='data',index=True,**kwargs):
        for para_i in self.paras_i(**kwargs):
            yield self.get_data(para_i,data_type=data_type,index=index,**self.kwargs(kwargs))
    def get_data(self,para_i,data_type='data',index=True,**kwargs):
        paras=[self.para(para_i,**kwargs)] if para_i else self.paras(**kwargs)
        o=[]
        for para in paras:
            if para is None: continue
            if hasattr(para,data_type):
                func=getattr(para,data_type)
                odf=func(index=index,**self.kwargs(kwargs))
                odf['para_i']=para.i
                o.append(odf)
        return concatt(o,index=index)

    #############################################
    ### Data types
    #############################################
    def words(self,para_i=None,**kwargs): return self.get_data(para_i,'words',**kwargs)
    @property
    def num_words(self): return sum(para.num_words for para in self.paras())
    
    def sents(self,para_i=None,**kwargs):
        sents=[sent for para in self.paras(para_i,**kwargs) for sent in para.sents(**kwargs)]
        return sents
    def sent(self,sent_i,para_i=None,**kwargs):
        if type(sent_i)==int:
            sents=self.sents(para_i,**kwargs)
            sent_i-=1
            if sent_i<len(sents): return sents[sent_i]
        elif type(sent_i) in {tuple,list}:
            para_i,sent_i=sent_i
            para=self.para(para_i,**kwargs)
            if para is not None:
                return para.sent(sent_i,**kwargs)
    @property
    def num_sents(self): return sum(para.num_sents for para in self.paras())
    @property
    def fsent_str(self): return self.para(1).fsent_str
    @property
    def lsent_str(self): return self.para(1).lsent_str

    def sylls(self,para_i=None,**kwargs): return self.get_data(para_i,'sylls',**kwargs)
    def syntax(self,para_i=None,**kwargs): return self.get_data(para_i,'syntax',**kwargs)
    def mtrees(self,para_i=None,**kwargs): return self.get_data(para_i,'mtrees',**kwargs)
    def data(self,para_i=None,**kwargs): return self.get_data(para_i,'data',**kwargs)
    def parse(self,para_i=None,**kwargs): return self.get_data(para_i,'parse',**kwargs)
    def parses(self,para_i=None,**kwargs): return self.get_data(para_i,'parses',**kwargs)
    def parse_iter(self,**kwargs):
        for para in self.paras(**kwargs):
            yield from para.parse_iter(**kwargs)



class ParaModel(TextModel):
    def __init__(self,txt,path_db=PATH_DB,i=None,**kwargs):
        self.txt=txt
        self._parses={}
        self._id=get_para_key(txt, kwargs)
        self._sents_str=tokenize_sents_txt(txt)
        self._words=None
        self._sylls=None
        self._doc=None
        self._syntax=None
        self._mtrees=None
        self._mtrees_obj=None
        self._path_db=path_db
        self._sents=None
        self._kwargs=kwargs
        self.i=i

    def __repr__(self):
        istr=f' #{self.i}' if self.i else ''
        return f'<Paragraph/Stanza{istr}: “{self.fsent_str}...” ({self.num_sents} sents, {self.num_words} words)>'.replace('\n',' ')

    

    ####################################
    ## WORDS
    ####################################

    def words(self,index=True,**kwargs):
        if self._words is None:
            iterr=tokenize_sentwords_iter(self.txt,sents=self._sents_str,**self.kwargs(kwargs))
            self._words=pd.DataFrame(iterr)
        return setindex(self._words) if index else resetindex(self._words)

    @property
    def num_words(self): return len(self.words(index=False))
    
    ####################################
    ## SENTS
    ####################################

    # def sent(self,sent_i):

    def sents(self,**kwargs):
        if self._sents is None:
            self._sents=[]
            doc = self.doc(**kwargs)
            sentobjs = list(doc.sentences) if doc is not None else []
            for si,sent_str in enumerate(self._sents_str):
                sent_i=si+1
                nlp_sent_obj = sentobjs[si] if si<len(sentobjs) else None
                sentobj = Sent(sent_str, i=sent_i, nlp_sent_obj=nlp_sent_obj, **self.kwargs(kwargs))
                self._sents.append(sentobj)
        return self._sents
    
    def sent(self,sent_i,**kwargs):
        sents=self.sents(**self.kwargs(kwargs))
        si=sent_i-1
        return sents[si] if si<len(sents) else None
    
    @property
    def num_sents(self): return len(self._sents_str)
    @property
    def fsent_str(self): return self._sents_str[0]
    @property
    def lsent_str(self): return self._sents_str[-1] if self.num_sents>1 else ''

    ####################################
    ## SYLLABIFY
    ####################################

    def syllabify(self,df,index=False,**kwargs):
        kwargs=self.kwargs(kwargs)
        odf=syllabify_df(df=df,**kwargs)
        return setindex(odf) if index else resetindex(odf)

    def sylls(self,df=None,index=True,**kwargs):
        if df is None: df=self.words(index=False)
        return self.syllabify(df,index=index,**self.kwargs(kwargs))

    ####################################
    ## NLP DOCS
    ####################################

    def doc(self,**kwargs):
        if self._doc is None:
            sentwords_ll=tokenize_sentwords_ll_from_df(self.words(index=False))
            self._doc=get_nlp_doc(sentwords_ll,**self.kwargs(kwargs))
        return self._doc

    ####################################
    ## NLP FEATS
    ####################################

    def syntax(self,index=True,**kwargs):
        if self._syntax is None:
            doc=self.doc()
            self._syntax=get_nlp_feats_df(doc, **self.kwargs(kwargs))
        return setindex(self._syntax) if index else resetindex(self._syntax)

    ####################################
    ## Mtrees
    ####################################

    def mtrees(self,index=True,**kwargs):
        if self._mtrees is None:
            o=[
                sent.mtree_df(**self.kwargs(kwargs)).assign(sent_i=sent.i)
                for sent in self.sents(**self.kwargs(kwargs))
            ]
            self._mtrees=pd.concat(o) if len(o) else pd.DataFrame()
        return setindex(self._mtrees) if index else resetindex(self._mtrees)

    ####################################
    ## Combined
    ####################################

    def data(self,
            syntax=True,
            sylls=True,
            mtree=True,
            index=True,
            save=True,
            **kwargs):

        odf = self.words(index=False,**self.kwargs(kwargs))
        if syntax: odf=safe_merge(odf,self.syntax(index=False,**self.kwargs(kwargs)))
        if mtree: odf=safe_merge(odf,self.mtrees(index=False,**self.kwargs(kwargs)))
        if sylls: odf=self.syllabify(df=odf,**self.kwargs(kwargs))
        if save: self.save()
        return setindex(odf) if index else resetindex(odf)


    ####################################
    ## Save
    ####################################

    def save(self,path_db=None):
        if path_db is None: path_db=self._path_db
        with dc.Cache(path_db) as db:
            #print(f'Saving {self._id}')
            db[self._id] = to_blosc(pickle.dumps(self))



    ####################################
    ## Units
    ####################################

    def iter_units(self,**kwargs):
        dfdata = self.data(index=False, sylls=True, **self.kwargs(kwargs))
        try:
            divide_parse_units(dfdata,**self.kwargs(kwargs))
            yield from (g for i,g in dfdata.groupby('unit_i'))
        except KeyError:
            yield pd.DataFrame()
    
    def units(self,index=True,**kwargs):
        o=list(self.iter_units(**self.kwargs(kwargs)))
        odf=pd.concat(o) if len(o) else pd.DataFrame()
        return setindex(odf) if index and len(odf) else resetindex(odf)

    ####################################
    ## Combos
    ####################################

    def iter_combos(self,**kwargs):
        for dfunit in self.iter_units(**self.kwargs(kwargs)):
            for dfcombo in iter_combos(dfunit,**self.kwargs(kwargs)):
                yield dfcombo

    ####################################
    ## Units
    ####################################
        
    def get_parse_key(self,kwargs,good_keys={'constraints'}):
        kwargs=dict((k,v) for k,v in kwargs.items() if k in good_keys)
        if not 'constraints' in kwargs or not kwargs['constraints']:
            kwargs['constraints']=DEFAULT_CONSTRAINTS
        else:
            kwargs['constraints']=[
                func if type(func)==str else func.__name__
                for func in kwargs['constraints']
            ]
        kwargs['constraints']=sorted(list(kwargs['constraints']))
        return kwargs_key(kwargs)


    def parse_iter(self,index=True,force=False,num_proc=None,incl_data=True,by_line=False,verbose=True,progress=True,**kwargs):
        if num_proc is None: num_proc = mp.cpu_count()//2
        key=self.get_parse_key(kwargs)
        if force or not key in self._parses:
            units=list(self.iter_units(**self.kwargs(kwargs)))
            num_units=len(units)
            oiterr=pmap_iter(
                parse_unit_combos,
                units,
                num_proc=num_proc,
                
                kwargs=self.kwargs(kwargs),
                progress=False#progress if not verbose else False
            )
        else:
            num_units=self._parses[key]['unit_i'].nunique()
            oiterr=(g for i,g in sorted(self._parses[key].groupby('unit_i')))
        o=[]

        oiterr=tqdm(
            oiterr,
            disable=not progress,
            desc='Metrically parsing line units',
            total=num_units,
            position=0
        )

        for odf in oiterr:
            if not len(odf): continue
            parse_str=odf.parse_str.iloc[0]
            unit_i=odf.unit_i.iloc[0]
            pref=f'[{unit_i}/{num_units}]'
            pref=format(STYLE_SANS, pref)
            ostr=f'{pref} {parse_str}'
            oiterr.set_description(ostr)
            if verbose and IS_INTERACTIVE: display(Markdown(parse_markdown(odf)))
            o.append(odf)
            if by_line: odf=to_lines(odf,**kwargs)
            odf=setindex(odf) if index else resetindex(odf)
            yield odf
        self._parses[key]=concatt(o)
    
    def parse(self,index=True,incl_data=True,force=False,by_line=False,verbose=True,**kwargs):
        key=self.get_parse_key(kwargs)
        parsed=False
        if force or not key in self._parses:
            o=[]
            kwargs=self.kwargs(kwargs)
            if 'by_line' in kwargs: del kwargs['by_line']
            kwargs['verbose']=verbose
            oiterr=self.parse_iter(force=force,by_line=False,**kwargs)
            list(oiterr)
            if not key in self._parses: return pd.DataFrame()
            parsed=True
        
        odf=self._parses[key]
        if not len(odf): return pd.DataFrame()

        if verbose and not parsed:
            for i,g in odf.groupby('unit_i'):
                printm(parse_markdown(g))

        if incl_data:
            data=self.data(**self.kwargs(kwargs))
            odf=safe_merge(odf, data, on=['sent_i','word_i','word_ipa_i','syll_i'],how='left')
        odf=setindex(odf) if index else resetindex(odf)
        if by_line: odf=to_lines(odf,**kwargs)
        return odf
        
    def parses(self,**kwargs): return self.parse(**kwargs)



class SentModel(ParaModel):
    def __init__(self,txt,i=None,nlp_sent_obj=None,**kwargs):
        self.txt=txt
        self.i=i
        self.nlp=nlp_sent_obj
        self._mtree=None
        self._mtree_df=None
        self._kwargs=kwargs

    def __repr__(self):
        txt=self.txt.replace("\n"," ")
        return f'<Sentence #{self.i}: {txt}>'
    
    def mtree(self,**kwargs):
        if self._mtree is None:
            self._mtree=get_mtree(self.nlp,**self.kwargs(kwargs))
        return self._mtree
    
    def mtree_df(self,**kwargs):
        if self._mtree_df is None:
            mtree=self.mtree(**self.kwargs(kwargs))
            self._mtree_df=mtree.get_stats(**self.kwargs(kwargs)) if mtree is not None else pd.DataFrame()
        return self._mtree_df#.assign(sent_i=self.i)
    
    def grid(self,**kwargs):
        import plotnine as p9
        p9.options.figure_size=(11,5)
        figdf=resetindex(self.mtree_df(**kwargs))
        figdf['prom_tstress']+=0.1
        figdf['prom_tstress']=figdf['prom_tstress'].fillna(0)
        return p9.ggplot(
            figdf,
            p9.aes(x='word_i',y='prom_tstress',label='word_str')
        ) + p9.geom_col(alpha=.25) + p9.geom_text(size=9,angle=45) + p9.theme_void()